%:-module(createProduct, [ground/2,intersectCfgWithReg/3,validPair/3]).

/** <createProduct> Create intersection of a control flow grammar with regular grammar
The grammars are represented as a set of rules.
The intersection is created in a symbolic manner. 
    
@author Tomas Poch
*/


%% intersectRules(+SUPPORTTOKEN,+CFRULES,-RESULTRULES)
% 
% Create grammar rules capturing the intersection of the given context-free grammar with a regular grammar
%
% The intersection is created in a symbolic manner.
% Non-terminals of the intersection grammar have the structure pr(CFNT,PREREG,POSTREG) where CFNT is the 
% non-terminal of the input cf grammar, PREREG is the non-terminal of the regular grammar before entering
% the CFNT while the POSTREG is the non-terminal of the regular grammar after leaving CFNT. 
%
% The result rules assume that the regular grammar is installed in the fact database at the time the result
% is used. Moreover, depending whether the groundRules/0 predicate holds, the non-terminals PREREG and POSTREG
% are forced to be ground. In such case the set of valid values is selected by the SUPPORTTOKEN term. 
% (To avoid mixing of memory and pattern non-terminal values) 
%
% @param SUPPORTTOKEN used to select grounded values of the regular non-terminals
% @param CFRULES list of input rules 
% @param RESULTRULES list of intersection rules
intersectRules(_,[],[]).
intersectRules(SUPPORTTOKEN,[H|T],[RESULT1|RESULT]):-
  intersectRules_hlp(SUPPORTTOKEN,H,RESULT1),
  intersectRules(SUPPORTTOKEN,T,RESULT).

intersectRules_hlp(SUPPORTTOKEN,(X:-CFGCOND),(RESULTRULE:-(CFGCOND,REGCOND))):-
  intersectRules_hlp(SUPPORTTOKEN,X,(RESULTRULE:-REGCOND)).

intersectRules_hlp(SUPPORTTOKEN,rule(X,eps),(rule(pr(X,D0,D0),eps):-ground(SUPPORTTOKEN,D0))).
intersectRules_hlp(SUPPORTTOKEN,rule(X,term(T)),(rule(pr(X,D0,D1),term(T)):-rule(D0,term(T),D1),validPair(SUPPORTTOKEN,D0,D1))).
intersectRules_hlp(SUPPORTTOKEN,rule(X,term(T),Y),(rule(pr(X,D0,D2),term(T),pr(Y,D1,D2)):-rule(D0,term(T),D1),ground(SUPPORTTOKEN,D0),ground(SUPPORTTOKEN,D1),ground(SUPPORTTOKEN,D2))):-isNonTerminal(Y).
intersectRules_hlp(SUPPORTTOKEN,rule(X,Y,Z),(rule(pr(X,D0,D2),pr(Y,D0,D1),pr(Z,D1,D2)):-validPair(SUPPORTTOKEN,D0,D1),validPair(SUPPORTTOKEN,D1,D2))):-isNonTerminal(Y),isNonTerminal(Z).
intersectRules_hlp(SUPPORTTOKEN,rule(X,Y),(rule(pr(X,D0,D1),pr(Y,D0,D1)):-validPair(SUPPORTTOKEN,D0,D1))):-isNonTerminal(Y).


%% intersectCfgWithReg(+CFGRM,+REGGRM,-RESGRM)
% 
% Create intersection of the given control-flow and regular gramar. 
% The implementation follows the algorithm suggested in Aho, Sethi, Ullmann: Compilers: Principles, Techniques, and Tools in a symbolic way.
% It can process also control-flow grammar originated as the result of this predicate.
%
% @param CFGRM input context-free grammar represented as the term cfGrm(INIT, CFRULES, SUPPORTRULES) - initial non-terminal, list of 
% grammar rules and list of rules used in the grammar rules bodies. 
% @param REGGRM regular grammar represented as the term regGrm(INIT,REGRULES,SUPPORTTOKEN) - initial non-terminal, list of rules and the term used to ground non-terminals. 
% @param RESGRM result context-free grammar. The list of support rules contains the original rules of the regular grammar.
%
intersectCfgWithReg(cfGrm(CFAXIOM,CFRULES),REGGRM,RESGRM):-
  intersectCfgWithReg(cfGrm(CFAXIOM,CFRULES,[]),REGGRM,RESGRM).

intersectCfgWithReg(cfGrm(CFAXIOM,CFRULES,SUPPORTRULES),
                    regGrm(REGINIT,REGRULES,SUPPORTTOKEN,encodeInfo(ENCODERULES,REGDOMAIN)),
                    cfGrm(RESULTAXIOM,[INITRULE|RESULT],[ENCODENORM,ENCODEINIT|SUPPORT2])):-
  intersectRules(SUPPORTTOKEN,CFRULES,RESULT),
  constructIntersectAxiom(SUPPORTTOKEN,CFAXIOM,REGINIT,INITRULE,RESULTAXIOM),!,
  ENCODENORM=(
    encodeNT(pr(PC,R1,R2),CODE):-
      encodeNT(PC,CODEPC),encodeNT(SUPPORTTOKEN,R1,CODER1), encodeNT(SUPPORTTOKEN,R2,CODER2),
      CODE is ((CODEPC*REGDOMAIN+CODER1)*REGDOMAIN+CODER2)*2
  ),
  ENCODEINIT=(
    encodeNT(init(PC,R1),CODE):-
      encodeNT(PC,CODEPC),encodeNT(SUPPORTTOKEN,R1,CODER1),
      CODE is (CODEPC*REGDOMAIN+CODER1)*REGDOMAIN*2+1
  ),
  append([SUPPORTRULES,REGRULES,ENCODERULES],SUPPORT2).

%intersectCfgWithPattern(CFGRM,REGGRM,cfGrm(init(CFINIT,p(0,0)),RULES,ENC)):-
%  intersectCfgWithReg(CFGRM,REGGRM,cfGrm(init(CFINIT,_),[rule(INIT,pr(INITVAR,|RULES],ENC)).

%intersectCfgWithPattern(CFGRM,
%                        regGrm(_,REGRULES,SUPPORTTOKEN,EI),
%			RESULT):-
%  intersectCfgWithReg(CFGRM,
%                      regGrm(p(0,0),REGRULES,SUPPORTTOKEN,EI),
%		      RESULT).



similarPC(pr(X,_,_),pr(X,_,_)):-!.
similarPC(init(X,_),init(X,_)):-!.
similarPC(z(X),z(X)):-!.
similarPC(X,X):-atom(X),!.
%similarPC(X,Y):-writeln(wronginit:(X:Y)),fail.

%% constructIntersectAxiom(+SUPPORTTOKEN,+CFINIT,+RINIT,-RULE,-INIT)
%
% Create initial non-terminal of the intersection grammar and the initial rule.
% The initial non-terminal takes the form init(CFINIT,RINIT) determining the initial non-terminals of both grammars and
% the initial rule transforms it into states taking the form pr(CFINIT,RINIT,RFINAL), where RFINAL enumerates all final
% non-terminals of the regular grammar (those that can be rewriten to empty symbol)
%
% @param SUPPORTTOKEN token used to ground the regular part of non-terminals
% @param CFINIT initial non-terminal of thecontext-free grammar
% @param RINIT  initial non-terminal of the regular grammar
% @param RULE  initial rule of the result
% @param INIT  initial non-terminal of rule of the result context-free grammar.
%
constructIntersectAxiom(SUPPORTTOKEN,CFINIT,RINIT,RULE,INIT):-
  RULE=(rule(init(INITVAR,RINIT),pr(INITVAR,RINIT,RFINAL)):-similarPC(INITVAR,CFINIT),ground(SUPPORTTOKEN,RINIT),ground(SUPPORTTOKEN,RFINAL),rule(RFINAL,eps)),
  INIT=init(CFINIT,RINIT).
  
%% validPair(+SUPPORTTOKEN,-D0,-D1)
% 
% Used by the intersection rules to ground pair of regular parts of non-terminals
% Depending on groundRules/0 , the parts are forced to be ground.
%
validPair(SUPPORTTOKEN,D0,D1):-
  ground(SUPPORTTOKEN,D0),ground(SUPPORTTOKEN,D1).

%% ground(+SUPPORTTOKEN,-P)
% 
% Used by the intersection rules to ground regular part of non-terminals
% Depending on groundRules/0 , the parts are forced to be ground.
%
ground(SUPPORTTOKEN,A):-
  groundRules->
    valid(SUPPORTTOKEN,A);
    checkForm(SUPPORTTOKEN,A).

checkForm(ptrn,p(_,_)).
checkForm(gmem,s(_,_)).
checkForm(gmem,errmem).

