/** <ctxSwitchRules> Add context switching rules into the context free grammar. 
The input grammar represents already both control-flow and memory.
Introduce an inactive NT for each NT of the input grammar. The transition rules
between active and inactive states produce yield/go terminals.   

@author Tomas Poch
*/

%:-module(ctxSwitchRules, [generateCtxRules/4]).

%% generateCtxRules(TID,TIDS,PROGRULES,RESULT)
%
% Generate context switching rules. It involves switching from active to inactive non-terminals (ctxSwitchRules/3)
% and terminating in some inactive non-terminals (inactiveEpsRules/2).
%
% @param TID identifier of this thread (used in go() terminals leading to active non-terminals)
% @param TIDS list of all thread identifiers (except for TID usedin go() terminals staying in an inactive non-terminal)
% @param PROGRULES list of original rules
% @param RESULT list of context switching rules. Does not include PROGRULES.
%
generateCtxRules(TID,TIDS,_,[(rule(pr(z(PC),errmem,errmem),eps):-validpc(PC)),(rule(pr(PC,errmem,errmem),eps):-validpc(PC))|RESULT]):-
  ctxSwitchRules(TID,TIDS,RESULT).

%generateCtxRules(TID,TIDS,PROGRULES,RESULT):-
%  ctxSwitchRules(TID,TIDS,SWRULES),
%  inactiveEpsRules(PROGRULES,EPSRULES),
%  append(SWRULES,EPSRULES,RESULT).



%% ctxSwitchRules(+TID,+TIDS,-RULES)
%
% Generate rules switching from active to inactive non-terminals.
%
% For each non-terminal NT=pr(PC,PRE,POST) create symbolic rule going to its inactive counterpart z(NT)
% while producing yield(PRE) terminal. In the non-active counter part allow accepting of other yields 
% terminals and change the non-terminal to reflect the last state of the global memory. Finally, at each
% inactive non-terminal allow accepting of the go(TID) terminal to go to the corresponding active non-terminal
% Allow also accepting of go(OTHERTID) - the non-terminal remains unchanged in this case.
%
% @param TID identifier of this thread (used in go() terminals leading to active non-terminals)
% @param TIDS list of all thread identifiers (except for TID usedin go() terminals staying in an inactive non-terminal)
% @param RULES list of context switching rules. 
%
ctxSwitchRules(TID,TIDS,RULES):-
  RULES=[
  (rule(pr(PC,PRE,POST),term(yield(GPRE)),pr(z(PC),PRE,POST)):-(validpc(PC),PC\=z(_),PRE=s(GPRE,_),valid(gmem,PRE),valid(gmem,POST))),
  (rule(pr(z(PC),s(GPRE,LPRE),POST),term(yield(GPRE2)),pr(z(PC),s(GPRE2,LPRE),POST)):-(validpc(PC),valid(gmem,s(GPRE,LPRE)),valid(gmem,s(GPRE2,LPRE)),valid(gmem,POST))),
  (rule(pr(PC,errmem,POST),term(yield(errmem)),pr(z(PC),errmem,POST)):-(validpc(PC),PC\=z(_),valid(gmem,POST))),
  (rule(pr(z(PC),s(GPRE,LPRE),POST),term(yield(errmem)),pr(z(PC),errmem,POST)):-(validpc(PC),valid(gmem,s(GPRE,LPRE)),valid(gmem,POST))),
  (rule(pr(z(PC),errmem,POST),term(yield(errmem)),pr(z(PC),errmem,POST))),
  (rule(pr(z(PC),PRE,POST),term(go(TID)),pr(PC,PRE,POST)):-(validpc(PC),valid(gmem,PRE),valid(gmem,POST))),
  (rule(pr(z(PC),PRE,POST),term(go(OTHERTID)),pr(z(PC),PRE,POST)):-(validpc(PC),member(OTHERTID,TIDS),OTHERTID\=TID,valid(gmem,PRE),valid(gmem,POST)))
  ].

%% inactiveEpsRules(+RULES,-RESULT).
% For each original non-terminal that rewrites to silent action allow also its inactive counterpart to do so. 
% The goal is to allow inactive thread to terminate (or to return from a function - no harm since it can not 
% continue until it get active)
%
% For each original non-terminal rewriting to 'final' non-terminal allow also its inactive counterpart to do so. 
% The 'final' nonterminal by convention denotes the terminal point of computation of assert threads. The same reasoning as for silent actions.
%
% @param TID identifier of this thread (used in go() terminals leading to active non-terminals)
% @param TIDS list of all thread identifiers (except for TID usedin go() terminals staying in an inactive non-terminal)
% @param RULES list of original rules
% @param RESULT list of context switching rules. Does not include PROGRULES.
%
inactiveEpsRules([],[]).
inactiveEpsRules([H|T],R):-inactiveEpsRules(T,T2),inactiveEpsRule(H,HR),(HR=null->R=T2;R=[HR|T2]).

inactiveEpsRule((H:-B),R):-!,inactiveEpsRule(H,HR),(HR=null->R=null;R=(HR:-B)).
inactiveEpsRule(rule(pr(PC,VARS,VARS),eps),rule(pr(z(PC),VARS,VARS),eps)):-!.
inactiveEpsRule(rule(pr(PC,VARS,VARS),pr(final,VARS,VARS)),rule(pr(z(PC),VARS,VARS),pr(z(final),VARS,VARS))):-!.
inactiveEpsRule(_,null).

