/** <execute> Simulate the grammar derivation from the view of the original program. 
@author Tomas Poch
*/

%:-module(execute, [isNonTerminal/1,extractInitPcAndMem/4,extractPc/3]).


%% isNonTerminal(?NT)
%
%  Holds for symbols that are non-terminals
%  Terminals are either silent action or term/1 term.
isNonTerminal(NT):-NT\=term(_),NT\=eps.

%% nextStep(+NT,-CHOICE)
%
%  For the given non-terminal get an option provided by the grammar.
%  Assumes the grammar installed in the prolog database.
%
%  @param NT     non-terminal
%  @param CHOICE choice in form 'choice(TERMINAL,NONTERMINALS)' 
%                Terminal is a symbol that can be generated by the grammar if NT is at the top of the stack.
%                NONTERMINALS is a list of NONTERMINALS that replace NT on the top of the stack.
%
nextStep(NT,choice(term(T),[])):-rule(NT,term(T)).
nextStep(NT,choice(silent,[])):-rule(NT,eps).
nextStep(NT,choice(silent,[NEWNT])):-rule(NT,NEWNT),isNonTerminal(NEWNT).
nextStep(NT,choice(term(T),[NEWNT])):-rule(NT,term(T),NEWNT),isNonTerminal(NEWNT).
nextStep(NT,choice(silent,[NEWNT,NEWNT2])):-rule(NT,NEWNT,NEWNT2),isNonTerminal(NEWNT),isNonTerminal(NEWNT2).

%% extractPc(+MODE, +PRODUCTNT, -CFNT)
%
% From the given non-terminal of an intersection grammar, extract the non-terminal of original context-free grammar
%
% @param MODE Either alllevels or toplevel. The result differs when the grammar is result of several intersections.
%             For 'alllevels' extract the non-terminal of the original context-free grammar entering the first intersection
%             For 'toplevel'  ommit just the fragment created by the last intersection.
% @param PRODUCTNT non-terminal of intersection grammar
% @param CFNT context-free non-terminal
extractPc(_,PC,PC):-PC\=pr(_,_,_),PC\=eps.
extractPc(alllevels,pr(SUBGRM,_,_),PC):-extractPc(alllevels,SUBGRM,PC).
extractPc(toplevel,pr(PC,_,_),PC).

%% extractPreMem(+MODE, +PRODUCTNT, REGNTS)
%
% From the given non-terminal of an intersection grammar, extract the non-terminals of regular grammars.
% The intersection non-terminal consists from context-free non-terminal CFNT and tuple of regular non-terminals PRE and POST
% refering to the state before entering CFNT and after leaving CFNT. This predicate extract the PRE nonterminal for each regular
% grammar involved in the intersection
%
% @param MODE Either alllevels or toplevel. The result differs when the grammar is result of several intersections.
%             For 'alllevels' extract the PRE non-terminals of all regular grammars involved in the intersection.
%             For 'toplevel'  extract the PRE non-terminal of the last reguler grammar involved in the last intersection.
% @param PRODUCTNT non-terminal of intersection grammar
% @param REGNTS list of non-terminals belonging to regular grammars
extractPreMem(_,PC,[]):-PC\=pr(_,_,_).
extractPreMem(alllevels,pr(SUBGRM,MEM,_),[MEM|SUBMEM]):-extractPreMem(alllevels,SUBGRM,SUBMEM).
extractPreMem(toplevel,pr(_,MEM,_),[MEM]).

%% firstMem(+MODE,+INITMEM,+NTS,-REGNT).
%
% From the list of non-terminals of an intersection grammar, extract the regular non-terminal of the first item.
% There can be either 0,1 or two items in the list of non-terminals (new top of the stack). 
% For 0, the option represents return from the function. In such case the old content of memory is returned. (INITMEM).
% For 2, the option represents function call. In such case only the first item is interesting (the memory after the
% return from the function is yet to be determined.
%
% @param MODE Either alllevels or toplevel. 
% @param INITMEM  last content of memory
% @param NTS  list of non-terminals of intersection grammar
% @param CFNTS list of context-free non-terminals
firstMem(_,MEM,[],MEM).
firstMem(DEPTH,_,[H|_],MEM):-extractPreMem(DEPTH,H,MEM).

%% extractPcs(+MODE,+NTS,-CFNTS).
%
% From the given non-terminal of an intersection grammar, extract the non-terminal of original context-free grammar
%
% @param MODE Either alllevels or toplevel. 
% @param NTS  list of non-terminals of intersection grammar
% @param CFNTS list of context-free non-terminals
extractPcs(_,[],[]).
extractPcs(DEPTH,[H|T],[H2|T2]):-extractPc(DEPTH,H,H2),extractPcs(DEPTH,T,T2).

%% createCompositeState(+MODE,+CFNT,+REGNTS,COMPNT)
%
% From the non-terminals of the individual grammars compose state of the intersection gramar.
%
% @param MODE Either alllevels or toplevel. 
% @param CFNT context-free non-terminal
% @param REGNTS list of regular non-terminals
% @param COMPNT non-terminal of the intersection grammar.
createCompositeState(_,PC,[],PC).
createCompositeState(alllevels,PC,[MEM|SUBMEM],pr(SUBPC,MEM,_)):-createCompositeState(alllevels,PC,SUBMEM,SUBPC).
createCompositeState(toplevel,PC,[MEM],pr(PC,MEM,_)).


%% nextStepSimul(+MODE,+PREMEM,+PC,-CHOICE)
%
%  For the given program point (control-flow non-terminal) and content of memory produce an option (new program point and memory)
%  which can be done in the next step.
%  Assumes the grammar installed in the prolog database.
%
%  @param MODE   Either alllevels or toplevel.
%  @param PREMEM regular non-terminal
%  @param PC     context-free non-terminal representing program point
%  @param CHOICE choice in form 'choice(ACT,MEM,PCS)' 
%                ACT is a symbol that can be generated by the grammar in the next step
%                MEM is the next regular non-terminal
%                PCS is a list of NONTERMINALS that replace NT on the top of the stack.
%
nextStepSimul(DEPTH,PREMEM,PC,choice(ACT,MEM,NEWPC)):-
  createCompositeState(DEPTH,PC,PREMEM,COMPST),
  nextStep(COMPST,choice(ACT,NXT)),
  extractPcs(DEPTH,NXT,NEWPC),
  firstMem(DEPTH,PREMEM,NXT,MEM).

%% writeOptions(+IDX,+OPTLIST)
%
% Print out the options for the next step. One option per line, indexed
% @param IDX initial index
% @param OPTLIST list of options
%
writeOptions(_,[]).
writeOptions(IDX,[H|T]):-
  writeln(IDX=H),
  succ(IDX,IDX2),
  writeOptions(IDX2,T).

%% simulateStack(+MODE,+INITMEM,+STACK)
%
%  Simulate the derivation of the grammar. At each step ask the user which
%  rule to use. The options are presented from the perspective of the 
%  original program locations and content of memory.
%
%  @param MODE   Either alllevels or toplevel.
%  @param MEM    Initial state of memory
%  @param STACK  list of non-terminals representing the stack.
%
simulateStack(_,_,[]).
simulateStack(DEPTH,MEM,[H|T]):-
  notFailingSetOf(OPTION,nextStepSimul(DEPTH,MEM,H,OPTION),OPTIONS),
  writeln((pc=H,mem=MEM)),
  writeOptions(0,OPTIONS),
  read(ITEM),
  (number(ITEM)->
  (
  nth0(ITEM,OPTIONS,choice(_,NEWMEM,NEWTOP)),
  append(NEWTOP,T,NEWSTACK),
  simulateStack(DEPTH,NEWMEM,NEWSTACK)
  );
  (ITEM=cont->true;fail)
  ).

%% extractInitPcAndMem(+MODE,+INIT,-PC,-MEM)
%
%  Extract the context-free non-terminal and regular-nonterminal from the initial 
%  non-terminal of the intersection grammar
%
extractInitPcAndMem(_,init(PC,MEM),PC,[MEM]):-PC\=init(_,_).
extractInitPcAndMem(alllevels,init(PC,MEM),INIT,[MEM|SUBMEM]):-extractInitPcAndMem(alllevels,PC,INIT,SUBMEM).
extractInitPcAndMem(toplevel,init(PC,MEM),PC,[MEM]).

%% simulateGrammar(+MODE, +CFGRM)
%
%  Simulate the given intersection grammar from the perspective of original grammars (control flow and memory). 
%  Used for debuging to see what really happens in the grammar.
%
%  At each step asks the user what rule to use. The options are presented in the form of non-terminal of context-free grammar 
%  and list of non-terminals of regular grammars. Depending on the mode, either the context-free grammar represents the
%  original control-flow grammar (alllevels) or the context-free grammar entering the last intersection (toplevel)
%
%  @param MODE   Either alllevels or toplevel.
%  @param CFGRM  Grammar to simulate.
%
simulateGrammar(DEPTH,CFGRM):-
   installGrammar(CFGRM),
   (validRuleMap(noAbstAtAll,RULE)-> % false for the last grammar, which is always generated to ensure patterns with yield/go. We do not want to simulate the artificial grammar
     (assertz(RULE),
      CFGRM = cfGrm(INIT,_,_),
      extractInitPcAndMem(DEPTH,INIT,CFAXIOM,MEM),
      simulateStack(DEPTH,MEM,[CFAXIOM])
     );
     true 
   ).

