/** <generateEquations> Generate the Presburger formula from the given context-free grammar.

Assumes the grammar rules in the form irule(RULEINDEX,LHS,RHS [,RHS2] ) instantiated 
in the prolog database. These rules are already instantiated rules - no symbolic 
representation, all terms are ground. These rules are indexed.

@author Tomas Poch
*/

%:-module(generateEquations, [generateEquation/8,mergeFormulas/3,collectVarsFromMaps/2,conjunction/2]).


%% getRHSRuleFor(+SYMBOL,-RULEID,-LHS)
%  
%  Find a rule using the given symbol (non-terminal/terminal) on the right-hand side. Get the left-hand
%  side of the rule as well.
%
%  @param SYMBOL symbol of the grammar
%  @param RULEID the index of the rule using the SYMBOL on the right-hand side
%  @param LHS the left-hand side of the rule 
%
%   
getRHSRuleFor(RHS,RULEID,LHS):-
  irule(RULEID,LHS,RHS,_);irule(RULEID,LHS,_,RHS);irule(RULEID,LHS,RHS).


getRHSRuleFor(RHS,RULEID):-getRHSRuleFor(RHS,RULEID,_).

%% getLHSRuleFor(+SYMBOL,-RULEID)
%  
%  Find a rule using the given symbol (non-terminal/terminal) on the left-hand side. 
%
%  @param SYMBOL symbol of the grammar
%  @param RULEID the index of the rule using the SYMBOL on the left-hand side
%
%   
getLHSRuleFor(LHS,RULEID):-
  irule(RULEID,LHS,_,_);irule(RULEID,LHS,_).





%% createSumByIndexes(+IDXS,+RULEVARS,-SUM)
%
% Given the list of indexes into RULEVARS list containing all rule vars
% create the sum of indexed vars
%
% @param IDXS list of indexes into RULEVARS
% @param RULEVARS list of prolog variables representing rule vars
% @param SUM result sum
createSumByIndexes([H|T],RULEVARS,VAR+TEXPR):-T\=[],
    nth0(H,RULEVARS,VAR),
    createSumByIndexes(T,RULEVARS,TEXPR).

createSumByIndexes([H],RULEVARS,VAR):-
    nth0(H,RULEVARS,VAR).

createSumByIndexes([],_,0).


%% generateSumInput(+NONTERM,-EXPR,?RULEVARS)
%
%  For the given non-terminal create the sum of all rule variables leading to the non-terminal
%  
%  @param NONTERM  non-terminal appearing on the right hand side of all rules involved in the sum
%  @param EXPR     result expression
%  @param RULEVARS rulevars used in the expression
%
generateSumInput(NONTERM,EXPR,RULEVARS):-
  notFailingBagOf(RID,getRHSRuleFor(NONTERM,RID),RIDS),
  createSumByIndexes(RIDS,RULEVARS,EXPR).

%% generateSumOutput(+NONTERM,-EXPR,?RULEVARS)
%
%  For the given non-terminal create the sum of all rule variables having the non-terminal on their left hand side.
%  
%  @param NONTERM  non-terminal appearing on the left hand side of all rules involved in the sum
%  @param EXPR     result expression
%  @param RULEVARS rulevars used in the expression
%
generateSumOutput(NONTERM,EXPR,RULEVARS):-
  notFailingBagOf(RID,getLHSRuleFor(NONTERM,RID),RIDS),
  createSumByIndexes(RIDS,RULEVARS,EXPR).

%%  generateNonTerminalEq(+INIT,?RULEVARS,+NONTERMS,-EQS).
%
%  Generate the Kirchoff-like rule equation for each non-terminal.
%  Number of appearances on left hand side and right hand side of the rules used in the derivation is equal
%
%  @param INIT initial non-terminal its equation is a bit-different (+1)
%  @param RULEVARS rule variables counts the number of usages of each rule. Represented as prolog variables
%  @param NONTERMS list of non-terminals
%  @param EQS list of Kirchoff equations
%  
generateNonTerminalEq(_,_,[],[]).
generateNonTerminalEq(INIT,RULEVARS,[H|T],[HR|TR]):-
    generateNonTerminalEq1(INIT,RULEVARS,H,HR),
    generateNonTerminalEq(INIT,RULEVARS,T,TR).
generateNonTerminalEq1(INIT,RULEVARS,INIT,1 + SUMINP - SUMOUTP #= 0):-
    generateSumInput(INIT,SUMINP,RULEVARS),
    generateSumOutput(INIT,SUMOUTP,RULEVARS).
generateNonTerminalEq1(INIT,RULEVARS,H,SUMINP - SUMOUTP #=0):-H\=INIT,
    generateSumInput(H,SUMINP,RULEVARS),
    generateSumOutput(H,SUMOUTP,RULEVARS).


%%  generateTerminalEq(+TERMS,?TERMVARS,?RULEVARS,-EQS).
%
%  Generate the equations expressing the number of terminals appearing in the result word. It is equal
%  to the number of usages of rules having the terminal on their right hand side.
%
%  @param TERMS terminals
%  @param TERMVARS term variables counts the number of occurences of the terminal in the generated word
%  @param RULEVARS rule variables counts the number of usages of each rule. Represented as prolog variables
%  @param EQS generated equations
%  
generateTerminalEq([],[],_,[]).
generateTerminalEq([TERM|TERMTAIL],[TVAR|TERMVARS],RULEVARS,[HR|TR]):-
    generateTerminalEq1(TERM,TVAR,RULEVARS,HR),
    generateTerminalEq(TERMTAIL,TERMVARS,RULEVARS,TR).
generateTerminalEq1(TERM,TVAR,RULEVARS,TVAR #= SUMINP):-
    generateSumInput(TERM,SUMINP,RULEVARS).

%%  generateTermReachabilityEq(+TERMS,?TERMVARS,-REACHVARS,EQS)
%
%  Generate equations stating that the terminals used in the generated word have to be reachable 
%  from the initial non-terminal by the rules used in derivation.
%  The predicate generate a reachability variable for each terminal and the equation states that once
%  the terminal is used in its associated reachability variable must not be zero. The values of
%  reachability variables are defined by equations defined in generateReachabilityEq/7
%
%  @param TERMS terminals
%  @param TERMVARS term variables counts the number of occurences of the terminal in the generated word
%  @param REACHVARS map (assoctree) associating terminals to their respective reachability variables. 
%         The variables are created by this predicate as well as the mapping.
%  @param EQS generated equations
%
generateTermReachabilityEq([],[],EMPTY,[]):-empty_assoc(EMPTY).
generateTermReachabilityEq([TERM|T1],[TERMVAR|T2],ASSOC,[TERMVAR #= 0 #\/ TERMREACHVAR #> 0|T3]):-
    generateTermReachabilityEq(T1,T2,ASSOC2,T3),put_assoc(TERM,ASSOC2,TERMREACHVAR,ASSOC).
    
%% associateNonTermReachVars(+NONTERMS,+INASSOC,-OUTASSOC)
%
%  Associate a new reachability variable with each given non-terminal.
%
%  @param NONTERMS non-terminals
%  @param INASSOC original mapping of grammar symbols to rachability variables
%  @param OUTASSOC new mapping
associateNonTermReachVars([],E,E).
associateNonTermReachVars([NONTERM|T],ASSOCIN,ASSOCOUT):-
    associateNonTermReachVars(T,ASSOCIN,ASSOCTMP),put_assoc(NONTERM,ASSOCTMP,_,ASSOCOUT).

%% getAllLHSLeadingTo(+SYMBOL,-RULEIDANDLHSES)
%
%  Find all rules using the given symbol (non-terminal/terminal) on the right-hand side and their left-hand
%  sides
%
%  @param SYMBOL symbol of the grammar
%  @param RULEIDANDLHSES list of pairs (RULEINDEX,LHS)
%
getAllLHSLeadingTo(TKN,R):-
  notFailingBagOf((RID,LHS),getRHSRuleFor(TKN,RID,LHS),R).

%% createReachableExpr(+INIT,+RVAR,+PAIRS,+MAP,+RULEVARS,-FRM)
%
%  Create the formula defining the value of the given reachability variable assuming it is non-zero. It is disjunct of subformulas
%  There is one subformula for each rule having the symbol associated with the given reachability variable on its RHS.
%  The one of these rules is chosen and the value is set to be higher by one than the value of reachability variable of the LHS symbol.
%
%  @param INIT initial non-terminal of the grammar
%  @param RVAR reachability variable to be defined by the formula
%  @param PAIRS list of tuples (RULEID,LHS) identifying rules having the symbol associated with RVAR on its right hand side
%  @param MAP structure associating symbols with reachability variables
%  @param RULEVARS rule variables counts the number of usages of each rule. Represented as prolog variables
%  @param FRM result formula
%  
createReachableExpr(_,_,[],_,_,0#=1).
createReachableExpr(INIT,RVAR,[(RID,LHS)|PTAIL],MAP,RULEVARS,(RVAR #= LHSVAR+1 #/\ RULEVAR #>0 #/\ LHSVAR #>0  ) #\/ REST):-
    LHS\=INIT,
    %get_assoc(TOKEN,MAP,RVAR),
    get_assoc(LHS,MAP,LHSVAR),
    nth0(RID,RULEVARS,RULEVAR),
    createReachableExpr(INIT,RVAR,PTAIL,MAP,RULEVARS,REST).

createReachableExpr(INIT,RVAR,[(RID,INIT)|PTAIL],MAP,RULEVARS,(RVAR #= 1 #/\ RULEVAR #>0) #\/ REST):-
    %get_assoc(TOKEN,MAP,RVAR),
    nth0(RID,RULEVARS,RULEVAR),
    createReachableExpr(INIT,RVAR,PTAIL,MAP,RULEVARS,REST).


%%  generateReachabilityDisj(+INIT,+SYMBOLS,+MAP,+RULEVARS, -FORMS)
%
%  For each symbol, create the formula defining the value of the associated reachability variable. If the value is higher
%  than 0, it has to be related to the values of neighbour variables as described in createReachableExpr/6.
%
%  @param INIT initial non-terminal of the grammar
%  @param SYMBOLS symbols for to define reachability variables for 
%  @param MAP structure associating symbols with reachability variables
%  @param RULEVARS rule variables counts the number of usages of each rule. Represented as prolog variables
%  @param FORMS result formulas
generateReachabilityDisj(_,[],_,_,[]).
generateReachabilityDisj(INIT,[TERM|TAIL],MAP,RULEVARS,[(RVAR#>0)#==>(EQ)|EQT]):-
    generateReachabilityDisj(INIT,TAIL,MAP,RULEVARS,EQT),
    getAllLHSLeadingTo(TERM,PAIRS),
    %(PAIRS=[]->writeln(suspic:TERM);true),
    get_assoc(TERM,MAP,RVAR),
    createReachableExpr(INIT,RVAR,PAIRS,MAP,RULEVARS,EQ).
    
%% generateRuleVarReachVarRelation(+INIT,+IDX,+RULEVARS,+REACHVARS,-FORMULAS)
%
%  For each rule variable define an equation relating its value to the value of reachability variable 
%  associated to its left hand side non-terminal. In paraticular, if the rule is used in the derivation (
%  value of rule variable is nonzero) then the left-hand-side non-terminal must be reachable (associated
%  reachability value is nonzero as well
% 
%  @param INIT initial non-terminal of the grammar
%  @param IDX index of the rule variable.
%  @param RULEVARS rule variables counts the number of usages of each rule. Represented as prolog variables
%  @param REACHVARS structure associating symbols with reachability variables
%  @param FORMULAS result formulas
%  
generateRuleVarReachVarRelation(_,_,[],_,[]).
generateRuleVarReachVarRelation(INIT,IDX,[RULEVAR|RULEVARS],MAP,EQLIST):-
    succ(IDX,IDX2), %IDX2 is IDX+1,
    generateRuleVarReachVarRelation(INIT,IDX2,RULEVARS,MAP,EQLIST1),
    (irule(IDX,SOURCENTERM,_);irule(IDX,SOURCENTERM,_,_)),
    get_assoc(SOURCENTERM,MAP,REACHVAR),
    (SOURCENTERM=INIT->EQLIST=EQLIST1;EQLIST = [(RULEVAR#>0 #==> REACHVAR#>0)|EQLIST1])
    .

%generateReachVarsBoundList([],_,[]).   
%generateReachVarsBoundList([REACHVAR|T],BOUND,[REACHVAR#=<BOUND|RT]):-
%    generateReachVarsBoundList(T,BOUND,RT).
%
%generateReachVarsBound(MAP,EQ):-
%    assoc_to_values(MAP,L),
%    length(L,BOUND),
%    generateReachVarsBoundList(L,BOUND,EQ). 

%%  generateReachabilityEq(+INIT,+TERMS,+NONTERMS,+TERMVARS,+REACHVARS,+RULEVARS,-FORMULAS)
%
%   Generate the formula expressing the fact that each used symbol of the grammar
%   must be reachable by the set of used rules.
%
%  @param INIT initial non-terminal of the grammar
%  @param TERMS list of terminals
%  @param NONTERMS list of non-terminals
%  @param TERMVARS list of terminal variables counting the number of occurences of each terminal in the generated word.
%  @param RULEVARS rule variables counts the number of usages of each rule. Represented as prolog variables
%  @param REACHVARS structure associating symbols with reachability variables
%  @param FORMULAS result formulas
%   
generateReachabilityEq(INIT,TERMS,NONTERMS,TERMVARS,MAP,RULEVARS,EQ):-
    generateTermReachabilityEq(TERMS,TERMVARS,TERMREACHVARMAP,EQ1),
    associateNonTermReachVars(NONTERMS,TERMREACHVARMAP,MAP),
    generateReachabilityDisj(INIT,TERMS,MAP,RULEVARS,TEQ),
    generateReachabilityDisj(INIT,NONTERMS,MAP,RULEVARS,NTEQ),
    %generateReachVarsBound(MAP,BOUNDEQ),    
    append([EQ1,TEQ,NTEQ],EQ).
   
%% indexRulesAndGenerateRuleVars(+RULES,-INDEXEDRULES,-RULEVARS)
%
%  Take the rules in form rule/[2,3] and transform them into irule/[3,4]. The first argument of irule is unique 
%  index of the rule. Generate the rule variables - variables used to represent the number of usages of the rule
%  in formulas
%
%  @param RULES list of rules
%  @param INDEXEDRULES list of indexed rules
%  @param RULEVARS list of rule variables
%
indexRulesAndGenerateRuleVars(RULES,INDEXEDRULES,RULEVARS,MAXIDX):-
    indexRulesAndGenerateRuleVars(0,RULES,INDEXEDRULES,RULEVARS,MAXIDX).
    
indexRulesAndGenerateRuleVars(N,[],[],[],N).
indexRulesAndGenerateRuleVars(IDX,[H|T],[H2|T2],[_|T3],N):-
    createIndexRule(IDX,H,H2),IDX2 is IDX + 1,
    indexRulesAndGenerateRuleVars(IDX2,T,T2,T3,N).

createIndexRule(IDX,(H1:-B),(H2:-B)):-createIndexRule(IDX,H1,H2).
createIndexRule(IDX,rule(X,Y,Z),irule(IDX,X,Y,Z)).
createIndexRule(IDX,rule(X,Y),irule(IDX,X,Y)).


%% generateEquation(+INIT,+TERMS,+NTERMS,?TERMVARS,+RULES,-FORMS,-RULEVARS,-REACHVARS)
%
%  Generate the list of formulas. The conjunction of the formulas characterizes the Parikh image
%  of the given context-free grammar.
%
%  @param INIT initial non-terminal of the grammar
%  @param TERMS list of terminals
%  @param NTERMS list of non-terminals
%  @param TERMVARS list of terminal variables - represents the Parikh image (number of occurences of the corresponding terminals in a word generated by the grammar).
%                  Represented as a list of prolog variables passed from the calling predicate. These variables are to be shared by all threads
%  @param RULES list of grammar rules
%  @param FORMS list of formulas
%  @param RULEVARS list of generated variables representing number of usages of particular rules. 
%  @param REACHVARS structure associating symbols with reachability variables
%   
generateEquation(INIT,TERMS,NTERMS,TERMVARS,RULES,[INITRVAR#=0|EQLIST],RULEVARS,REACHTKMAP):-
    indexRulesAndGenerateRuleVars(RULES,IRULES,RULEVARS,MAXIDX),!,
    writeln(rulevars=MAXIDX),
    installGrammarFacts(IRULES),
    %writeList(IRULES),
    generateNonTerminalEq(INIT,RULEVARS,NTERMS,NEQ),!,
    generateTerminalEq(TERMS,TERMVARS,RULEVARS,TEQ),!,
    append(NEQ,TEQ,EQLIST1),
    generateReachabilityEq(INIT,TERMS,NTERMS,TERMVARS,REACHTKMAP,RULEVARS,REQ),!,
    append(EQLIST1,REQ,EQLIST2),
    generateRuleVarReachVarRelation(INIT,0,RULEVARS,REACHTKMAP,RULEREACHRELATIONEQ),!,
    append(EQLIST2,RULEREACHRELATIONEQ,EQLIST),
    (get_assoc(INIT,REACHTKMAP,INITRVAR1)->INITRVAR=INITRVAR1;INITRVAR=1,writeln('Warning-axiom is not generating')).


%% conjunction(+FORMULALIST,-CONJUNCTION)
%
%  Create conjunction of the given formula list
%  
%  @param FORMULALIST either list of formulas or from(FORMLIST,_,_,_) term
%  @param CONJUNCTION conjunction of input formulas
conjunction([],true).
conjunction([H|T],RESULT):-conjunction(T,H,RESULT).
conjunction(form(EQLIST,_,_,_),RESULT):-conjunction(EQLIST,RESULT).
conjunction([H|T],RESULT,ACC):-conjunction(T,RESULT #/\ H,ACC).
conjunction([],LAST,LAST).

%% collectVarsFromMaps(REACHVARS,REACHVARSLIST)
%
%  Collect reachability variables from the tree
%
%  @param REACHVARS structure associating symbols with reachability variables
%  @param REACHVARSLIST reachability variables
%
collectVarsFromMaps([],[]).
collectVarsFromMaps([RH|RT],VARS):-
  collectVarsFromMaps(RT,VARS2),
  assoc_to_values(RH,VARS1),
  append(VARS1,VARS2,VARS).

%% mergeFormulas(+F1,+F2,-F)
%
%  Formulas are represented by the term form(FORMLIST, TVARS, RULEVARS, REACHVARS).
%  The predicate unifies terminal variables (TVAR) and merges the other lists.
%
%  @param F1 formula
%  @param F2 formula
%  @param F conjunction of F1 and F2
mergeFormulas(
    form(LIST1,TVARS,RULEVARS1,REACHMAPS1),
    form(LIST2,TVARS,RULEVARS2,REACHMAPS2),
    form(LIST,TVARS,RULEVARS,REACHMAPS)
    ):-
  append(LIST1,LIST2,LIST),
  append(RULEVARS1,RULEVARS2,RULEVARS),
  append(REACHMAPS1,REACHMAPS2,REACHMAPS).


