/** <instantiateRules> Transform the symbolic grammar into grounded grammar. 
 * All rules in the result use grounded non-terminals while generating the same language as the input symbolic grammar.
 
   @author Tomas Poch
 */

%:-module(instantiateRules, [insertListIntoSet/3,instantiateRules/4]).

%% instantiateRules(+CFGRM,+TERMS,-NONTERMS,-INSTANTIATEDRULES)
%
%  Instantiate the given symbolic grammar. 
%
%  @param CFGRM input grammar consisting of symbolic rules
%  @param TERMS set of terminals produced by the grammar
%  @param NONTERMS set of instantiated nonterminals actively used by the grammar 
%  @param INSTANTIATEDRULES set of instantiated rules

instantiateRules(cfGrm(INIT,SYMBOLICRULES,SUPPORTRULES),TERMS,ACTIVENTERMS,INSTANTIATEDRULES):-
   %writeList(SYMBOLICRULES),
   get_time(TS1),
   %instantiate the symbolic grammar. The rule/(2,3) predicates are used to traverse the 
   %grammar and obtain the approximation of reachable and generating non-terminals.
   installGrammar(cfGrm(INIT,SYMBOLICRULES,SUPPORTRULES)),
   %Ensure that the rule predicate unifies all variables with grounded terms
   %alternative algorithm based on simulation - not ready yet
   setRuleGrounding, 
   generateNTSetByCoreachability(cfGrm(INIT,SYMBOLICRULES,SUPPORTRULES),TERMS,ACTIVENTERMS,INSTANTIATEDRULES),
   % alternatively, one can run the following predicate to get the set of non-terminals actively used by the grammar
   %generateNTSetBySimulation(INIT,SYMBOLICRULES,ACTIVENTERMS,INSTANTIATEDRULES),TERMS=TERMS,
   %(verbose->write('\nReally used nonterminals:'),length(ACTIVENTERMS,CT),writeln(CT),writeList(ACTIVENTERMS);true),
   %(verbose->write('\nInstantiated rules:'),length(INSTANTIATEDRULES,COUNT),writeln(COUNT),writeRules(INSTANTIATEDRULES);true),
   
   (verbose->length(ACTIVENTERMS,NTERMSNUMBER),writeln(reachable_and_generating_set=NTERMSNUMBER),writeGrammarStat(INSTANTIATEDRULES);true),
   get_time(TS2),
   DT is TS2-TS1,
   (printTime->write(','),write(instantiationtime=DT);true),
   nl.

epsrule(rule(_,eps)).
commepsrule(rule(_,term(_))).
moverule(rule(_,Y)):-Y\=eps,Y\=term(_).
commrule(rule(_,term(_),_)).
fcncallrule(rule(_,Y,Z)):-Y\=term(_),Z\=term(_).


writeGrammarStat(INSTANTIATEDRULES):-
  partition(epsrule,INSTANTIATEDRULES,EPS_RULES,OTHER1),
  partition(commepsrule,OTHER1,COM_EPS_RULES,OTHER2),
  partition(moverule,OTHER2,MOVE_RULES,OTHER3),
  partition(commrule,OTHER3,COMM_RULES,OTHER4),
  partition(fcncallrule,OTHER4,FCNCALL_RULES,OTHER5),
  length(EPS_RULES,NOF_EPS_RULES),
  length(COM_EPS_RULES,NOF_COM_EPS_RULES),
  length(MOVE_RULES,NOF_MOVE_RULES),
  length(COMM_RULES,NOF_COMM_RULES),
  length(FCNCALL_RULES,NOF_FCNCALL_RULES),
  length(OTHER5,NOF_REST),
  writeln('Grammar stats'),
  writeln('X->eps':NOF_EPS_RULES),
  writeln('X->t':NOF_COM_EPS_RULES),
  writeln('X->Y':NOF_MOVE_RULES),
  writeln('X->tY':NOF_COMM_RULES),
  writeln('X->YZ':NOF_FCNCALL_RULES),
  writeln('Other':NOF_REST).



%% generateNTSetByCoreachability(+CFGRM,+TERMS,-NONTERMS,-INSTANTIATEDRULES)
%
%  Get the set of non-terminals actively used by the grammar when generating words.
%  The algorithm created the set of non-terminals that are both reachable and generating.
%
%  @param CFGRM input grammar consisting of symbolic rules
%  @param TERMS set of terminals produced by the grammar
%  @param NONTERMS set of instantiated nonterminals actively used by the grammar 
%  @param INSTANTIATEDRULES set of instantiated rules
generateNTSetByCoreachability(cfGrm(INIT,SYMBOLICRULES,SUPPORTRULES),TERMS,ACTIVENTERMS,INSTANTIATEDRULES2):-
   %get the approximation (superset) of grounded non-terminals that are both reachable and generating.
   %the result set is superset because of function call rules. The result contains grounded non-terminals 
   %for all possible function results given by the variable domains. The function body is not analyzed. 
   getActiveNTSuperset(cfGrm(INIT,SYMBOLICRULES,SUPPORTRULES),TERMS,ACTIVENTSUPERSET),
   %generate and instantiate rules capturing the superset. For each grounded non-terminal nt, predicate isActiveNT(nt) is produced.
   %
   %Instantiate symbolic rules. Uses the isActiveNT predicate to ground all grammar rules.
   generateRulePredicates(ACTIVENTSUPERSET,SYMBOLICRULES,INSTANTIATEDRULES),
   %Go through the set of grounded grammar rules and remove all rules that have at left-hand side a 
   %non-termnal which is not the initial one and does not appear at right side of any rule. This is 
   %done iteratively as long as the set of rules gets smaller. From this point ACTIVENTERMS contains
   %precisely the set of reachable and generating grounded non-terminals
   pruneUnreachableRules(INIT,INSTANTIATEDRULES,INSTANTIATEDRULES2,ACTIVENTERMS).




%% pruneUnreachableRules(+INIT,+INSTANTIATEDRULESIN,-INSTANTIATEDRULESOUT,-ACTIVENTERMS)
%
% Remove the unreachable rules - the rules such that the left-hand-side non-terminal is not reachable from the initial non-terminal
% 
% @param INIT initial non-terminal
% @param INSTANTIATEDRULESIN input list of rules
% @param INSTANTIATEDRULESOUT result list of rules
% @param ACTIVENTERMS list of non-terminals used in the result rules
%
pruneUnreachableRules(INIT,INSTANTIATEDRULESIN,INSTANTIATEDRULESOUT,ACTIVENTERMS):-
   getRHSNonTerminals(INSTANTIATEDRULESIN,RHSNTERMS),
   ACTIVENTERMS1=[INIT|RHSNTERMS],
   removeOrphanRules(INSTANTIATEDRULESIN,INSTANTIATEDRULESOUT1,ACTIVENTERMS1,REMOVED),
   (REMOVED=0->
     INSTANTIATEDRULESOUT=INSTANTIATEDRULESOUT1,ACTIVENTERMS=ACTIVENTERMS1
   ;
     pruneUnreachableRules(INIT,INSTANTIATEDRULESOUT1,INSTANTIATEDRULESOUT,ACTIVENTERMS)
   ).

%% removeOrphanRules(INSTANTIATEDRULESIN,INSTANTIATEDRULESOUT,NTERMS,REMOVED)
%
% Remove the rules such that the left-hand-side non-terminal neither occurs on a right-hand-side of a rule nor is the initial non-terminal.
%
% @param INSTANTIATEDRULESIN input list of rules
% @param INSTANTIATEDRULESOUT result list of rules
% @param NTERMS list of non-terminals used in the result rules
% @param REMOVED number of the rules removed (size of INSTANTIATEDRULESOUT minus size of INSTANTIATEDRULESIN)
%
removeOrphanRules(INSTANTIATEDRULES,INSTANTIATEDRULES2,NTERMS,REMOVED):-  
  emptySet(rr,E),insertListIntoSet(NTERMS,E,SET),
  removeOrphanRules1(INSTANTIATEDRULES,INSTANTIATEDRULES2,SET,REMOVED)
  %,writeln('nterms:'),writeList(NTERMS),writeln('------------'),writeln('before:'),writeList(INSTANTIATEDRULES),writeln('==========='),writeln('after:'),writeList(INSTANTIATEDRULES2)
  .
removeOrphanRules1([],[],_,0).
removeOrphanRules1([rule(LHS,A)|T],[rule(LHS,A)|T2],SET,R):-inSet(LHS,SET),!,removeOrphanRules1(T,T2,SET,R).
removeOrphanRules1([rule(LHS,A,B)|T],[rule(LHS,A,B)|T2],SET,R):-inSet(LHS,SET),!,removeOrphanRules1(T,T2,SET,R).
removeOrphanRules1([_|T],T2,SET,R):-removeOrphanRules1(T,T2,SET,R1),succ(R1,R).



% for debugging and verbose output
writeRules([]).
writeRules([H|T]):-writeRule(H),writeRules(T).
writeRule(rule(X,Y)):-write(X),write('\t->'),writeln(Y).
writeRule(rule(X,Y,Z)):-write(X),write('\t--'),write(Y),write('->\t'),writeln(Z).


%% getActiveNTSuperset(+CFGRM,+TERMS,-ACTIVENONTERMS)
%  
%  Generate all nonterminals of the given grammar that are reachable and generating
%  Depending on the tool setting, abstraction phase may be employed
%
%  Installs the symbolic rules and runs getActiveNT_hlp/4 to traverse the grammar 
%
%  @param CFGRM symbolic representation of the grammar
%  @param TERMS list of terminals generated by the grammar
%  @param ACTIVENONTERMS list of non-terminals that are both generating and reachable
%
%
getActiveNTSuperset(cfGrm(INIT,SYMBOLICRULES,SUPPORTRULES),TERMS,ACTIVENONTERMS):-
   (debug->writeList(SUPPORTRULES),writeList(SYMBOLICRULES),nl;true),
   getMaxAbstractInfo(INIT,MAXABSTINFO),
   useAbstraction(ABSTMODE),
   ONEMORERULE=(rule(pr(init(PC,MIN),P1,P2),pr(pr(PC,MIN,MOUT),P1,P2)):-
             (ground(gmem,MIN),ground(gmem,MOUT),rule(MOUT,eps)),validPair(ptrn,P1,P2)),
   assertz(ONEMORERULE),
   initialAbstInfo(MAXABSTINFO,ABSTMODE,ABSTINFO),
   getActiveNTSS_hlp(ABSTINFO,MAXABSTINFO,[abstRule(_,_,_),abstRule(_,_)],INIT,[ONEMORERULE|SYMBOLICRULES],TERMS,ACTIVENONTERMS)
   %getActiveNTSS_hlp(reachableAndGenerating,INIT,TERMS,ACTIVENONTERMS)
   %getActiveNT_hlp(reachable,INIT,TERMS,ACTIVENONTERMS)
   .

initialAbstInfo(noAbstAtAll,_,noAbstAtAll):-!.
initialAbstInfo(_,all,abstInfo(0,0)):-!.
initialAbstInfo(abstInfo(MAXG,_),local,abstInfo(MAXG,0)):-!.
initialAbstInfo(MAXABSTINFO,_,MAXABSTINFO).


%% getActiveNT_hlp(+MODE,+INIT,+TERMS,-ACTIVENONTERMS)
%  
%  Assuming the set of symbolic rules is installed in the prolog predicate database, it 
%  traverses it in order to get the set of non-terminals that are both reachable and generating.
%
%  @param MODE one of 
%    * generatingAndReachable - first get the set of generating nonterminals, then narrow it to contain only the nonterminals reachable from initial nonterminal 
%    * reachable - get the set of reachable nonterminals
%    * reachableAndGenerating - first get the reachable nonterminals, then narrow it to generating.
%  @param INIT the initial non-terminal
%  @param TERMS list of terminals generated by the grammar
%  @param ACTIVENONTERMS result list of non-terminals
%
getActiveNTSS_hlp(ABSTINFO,MAXAI,ABSTRACTRULES,INIT,SYMBOLICRULES,TERMS,NONTERMS):-
   (verbose->writeln('Abstraction phase':ABSTINFO);true),
   setCurrentAbstInfo(ABSTINFO),retractall(valid(gmem,s(_,_))),(validRuleMap(ABSTINFO,VALIDRULE)->assertz(VALIDRULE);true),
   getAbstract(ABSTINFO,INIT,ABSTRACTINIT),writeln(ABSTRACTINIT),   
   %(verbose->write('Searching for generating nonterminals, superset size:'),setSize(SUPERSET,SSL),writeln(SSL);true),
   retractall(abstRule(_,_)),retractall(abstRule(_,_,_)),installfacts(ABSTRACTRULES),length(ABSTRACTRULES,ARL),writeln(abstrules:ARL),%writeList(ABSTRACTRULES),
   getGenerating(TERMS,setContainingEverything,PHASE1),!,
   %getReachable(ABSTRACTINIT,setContainingEverything,PHASE1),
   (verbose->write('Searching for reachable nonterminals, generating size (including terminals):'),setSize(PHASE1,P1LT),writeln(P1LT);true),
   getReachable(ABSTRACTINIT,PHASE1,PHASE2),!,
   %getGenerating(TERMS,PHASE1,PHASE2),!,
   setToList(PHASE2,P2L),keepOnlyNonTerminals(P2L,NONTERMS2,_),

   %getGenerating(TERMS,PHASE1,PHASE2),
   (verbose->
     setToList(PHASE1,P1L),setToList(PHASE2,P2L),keepOnlyNonTerminals(P2L,MYNONTERMS,_),
     length(P1L,P1LL),length(MYNONTERMS,GLL),
     writeln((generating=P1LL,reachable=GLL))
   ;true
   ),
   (debug->writeln(generating),writeList(P1L),nl,writeln(generatingandreach),writeList(MYNONTERMS),nl;true),
   (getNextAbstractInfo(ABSTINFO,MAXAI,NEXTAI)->
     generateAbstractRulePredicates(NONTERMS2,SYMBOLICRULES,ABSTRACTRULES2),
     getActiveNTSS_hlp(NEXTAI,MAXAI,ABSTRACTRULES2,INIT,SYMBOLICRULES,TERMS,NONTERMS)
     ;
     NONTERMS=NONTERMS2
   ).




%% writePcList(+NONTERMS)
%  
%  Extract the non-terminals of the original control-flow grammar and print them out
%
%  @param NONTERMS list of non-terminals of the intersection grammar
%
writePcList(L):-writePcList_hlp(L,L2),sort(L2,L3),writeln(L3).

writePcList_hlp1(H,H2):-extractPc(alllevels,H,H2),!.
writePcList_hlp1(H,H2):-extractInitPcAndMem(alllevels,H,H2,_),!.
writePcList_hlp1(eps,eps).

writePcList_hlp([],[]).
writePcList_hlp([H|T],[H2|T2]):-writePcList_hlp(T,T2),writePcList_hlp1(H,H2).

%% createList(+SIZE,+ITEM,-LIST)
%
%  Generate the list containing SIZE ITEMS.
%  
%  @param SIZE number of items
%  @param ITEM item of the list
%  @param LIST result list
%
replaceInList(all,IN,_,IN):-!.
replaceInList(0,[],_,[]).
replaceInList(0,[_|INL],ITEM,[ITEM|OUTL]):-replaceInList(0,INL,ITEM,OUTL).
replaceInList(KEEP,[VAR|INL],ITEM,[VAR|OUTL]):-
   succ(N1,KEEP),
   replaceInList(N1,INL,ITEM,OUTL).

%%  getAbstractLoc(+LOC,-ABSTLOC)
%   
% Given a term representing local memory, generate its abstract version (all items are replaced by the top terminal) 
%
% @param LOC local memory term
% @param ABSTLOC abstract local memory term
% 
getAbstractPart(KEEPITEMS,VARTERM,ABSTVARTERM):-
   VARTERM=..[TYPE|INLIST],
   replaceInList(KEEPITEMS,INLIST,top,OUTLST),!,
   ABSTVARTERM=..[TYPE|OUTLST]
   .

%% getAbstract(+NT,-ABSTNT)
%
% Given the nonterminal, generate its abstract version (replace certain variables  by top)
% 
% @param NT input non-terminal
% @param ABSTNT output abstract non-terminal
%

getAbstract(abstInfo(GAI,LAI),init(init(PC,s(GLOB,LOC)),P0),init(init(PC,s(ABSTGLOB,ABSTLOC)),P0)):-
   !,getAbstractPart(LAI,LOC,ABSTLOC),getAbstractPart(GAI,GLOB,ABSTGLOB).

getAbstract(abstInfo(GAI,LAI),pr(init(PC,s(GLOB,LOC)),P0,P1),pr(init(PC,s(ABSTGLOB,ABSTLOC)),P0,P1)):-
   !,getAbstractPart(LAI,LOC,ABSTLOC),getAbstractPart(GAI,GLOB,ABSTGLOB).

getAbstract(abstInfo(GAI,LAI),init(pr(PC,s(GLOB1,LOC1),s(GLOB2,LOC2)),P0),init(pr(PC,s(ABSTGLOB1,ABSTLOC1),s(ABSTGLOB2,ABSTLOC2)),P0)):-
   !,getAbstractPart(LAI,LOC1,ABSTLOC1),
   getAbstractPart(LAI,LOC2,ABSTLOC2),
   getAbstractPart(GAI,GLOB1,ABSTGLOB1),
   getAbstractPart(GAI,GLOB2,ABSTGLOB2).

getAbstract(abstInfo(GAI,LAI),pr(pr(PC,s(GLOB1,LOC1),s(GLOB2,LOC2)),P0,P1),pr(pr(PC,s(ABSTGLOB1,ABSTLOC1),s(ABSTGLOB2,ABSTLOC2)),P0,P1)):-
   !,getAbstractPart(LAI,LOC1,ABSTLOC1),
   getAbstractPart(LAI,LOC2,ABSTLOC2),
   getAbstractPart(GAI,GLOB1,ABSTGLOB1),
   getAbstractPart(GAI,GLOB2,ABSTGLOB2). 

getAbstract(_,init(pr(PC,errmem,errmem),P0),init(pr(PC,errmem,errmem),P0)):-!.
getAbstract(_,pr(init(PC,errmem),P0,P1),pr(init(PC,errmem),P0,P1)):-!.

getAbstract(_,pr(pr(PC,errmem,errmem),P0,P1),pr(pr(PC,errmem,errmem),P0,P1)):-!.
getAbstract(abstInfo(GAI,LAI),pr(pr(PC,s(GLOB1,LOC1),errmem),P0,P1),pr(pr(PC,s(ABSTGLOB1,ABSTLOC1),errmem),P0,P1)):-
   !,getAbstractPart(LAI,LOC1,ABSTLOC1),getAbstractPart(GAI,GLOB1,ABSTGLOB1).

getAbstract(abstInfo(GAI,LAI),init(pr(PC,s(GLOB1,LOC1),errmem),P0),init(pr(PC,s(ABSTGLOB1,ABSTLOC1),errmem),P0)):-
   !,getAbstractPart(LAI,LOC1,ABSTLOC1),getAbstractPart(GAI,GLOB1,ABSTGLOB1).

getAbstract(noAbstAtAll,INIT,INIT).
:-multifile(inSet/2).
inSet(H,abstractSet(AI,ASSOC)):-!,getAbstract(AI,H,ABSTH),inSet(ABSTH,ASSOC).

%% getNextAbstractInfo(AIIN,INIT,AIOUT)
%

getMaxAbstractInfo(init(init(_,s(GLOB,LOC)),_),abstInfo(GSIZE,LSIZE)):-!,
  functor(GLOB,g,GSIZE),
  functor(LOC,l,LSIZE).

getMaxAbstractInfo(memDef(GLOBALS,LOCALS),abstInfo(GSIZE,LSIZE)):-!,
  length(GLOBALS,GSIZE),
  length(LOCALS,LSIZE).

getMaxAbstractInfo(_,noAbstAtAll).

getNextAbstractInfo(AI,AI,_):-!,fail.
getNextAbstractInfo(_,noAbstAtAll,_):-!,fail.
getNextAbstractInfo(abstInfo(MAXGAI,LAI),abstInfo(MAXGAI,_),abstInfo(MAXGAI,LAIOUT)):-!,succ(LAI,LAIOUT).
getNextAbstractInfo(abstInfo(GAI,0),_,abstInfo(GAIOUT,0)):-!,succ(GAI,GAIOUT).


%% keepOnlyNonTerminals(+SYMBOLS,-NTERMS,-TERMS)
%
%  Separate terminals from non-terminals
%
%  @param SYMBOLS input list of grammar symbols
%  @param NTERMS list of non-terminals from SYMBOLS
%  @param TERMS list of terminals from SYMBOLS
%
keepOnlyNonTerminals([],[],[]).
keepOnlyNonTerminals([term(X)|T],T1,[term(X)|T2]):-keepOnlyNonTerminals(T,T1,T2).
keepOnlyNonTerminals([eps|T],T1,T2):-keepOnlyNonTerminals(T,T1,T2).
keepOnlyNonTerminals([H|T],[H|T1],T2):-H\=term(_),H\=eps,keepOnlyNonTerminals(T,T1,T2).

%%  printGrammarStats(+SYMRULES,+INSTRULES,+SUPPORTRULES)
%
%  Printout grammar statistics 
%
%  @param SYMRULES list of symbolic rules
%  @param INSTRULES list of instantiated rules
%  @param SUPPORTRULES list of rules used in bodies of SYMRULES
%
printGrammarStats(SYMRULES,INSTRULES,SUPPORTRULES):-
  length(SYMRULES,NSYMRULES),
  length(INSTRULES,NINSTRULES),
  length(SUPPORTRULES,NSUPPORT),
  write(';'),write((symrules=NSYMRULES,instrules=NINSTRULES,supportrules=NSUPPORT)),nl.


writeInstRuleInfo(H,OUTR):-
  length(OUTR,OUTRSIZE),
  (OUTRSIZE>0->
    write(symrule=H),nl,write(instantiated=OUTRSIZE),nl,
    write('----------'),nl,writeList(OUTR),
    write('=========================================================='),nl
    ;
    true
  ).




%% getTerminalsAndNonterminals(RULES,TERMS,NTERMS)
%
% 
getTerminalsAndNonterminals(RULES,TERMS,NTERMS):-
  getTerminalsAndNonterminals_h(RULES,TERMSL,NTERMSL),
  delete(NTERMSL,eps,NTERMSL1),
  sort(TERMSL,TERMS),
  sort(NTERMSL1,NTERMS).


getTerminalsAndNonterminals_h([],[],[]).
getTerminalsAndNonterminals_h([H|T],TERMS,NTERMS):-
  getTerminalsAndNonterminals1(H,TERMSH,NTERMSH),
  getTerminalsAndNonterminals_h(T,TERMST,NTERMST),
  append(TERMST,TERMSH,TERMS),
  append(NTERMST,NTERMSH,NTERMS).


getTerminalsAndNonterminals1((H:-_),TERMS,NTERMS):-getTerminalsAndNonterminals1(H,TERMS,NTERMS).
getTerminalsAndNonterminals1(rule(NT,term(A)),[term(A)],[NT]).
getTerminalsAndNonterminals1(rule(NT,A),[],[A,NT]):-A\=term(_).

getTerminalsAndNonterminals1(rule(NT,term(A),term(B)),[term(A),term(B)],[NT]).
getTerminalsAndNonterminals1(rule(NT,term(A),B),[term(A)],[B,NT]):-B\=term(_).
getTerminalsAndNonterminals1(rule(NT,term(A)),[term(A)],[NT]).
getTerminalsAndNonterminals1(rule(NT,A,term(B)),[term(B)],[A,NT]):-A\=term(_).
getTerminalsAndNonterminals1(rule(NT,A,B),[],[A,B,NT]):-A\=term(_),B\=term(_).
getTerminalsAndNonterminals1(valid(_,_),[],[]).

notARegFragment(X):-X\=s(_,_),X\=p(_,_),X\=errmem.

encode(eps,-1):-!.
encode(term(ptrnWord(N)),MN):-!,MN is (0-N)-2.
encode(NT,CODE):-encodeNT(NT,CODE).


:-multifile(inSet/2).
inSet(X,setContainingEverything):-notARegFragment(X).

setSize(setContainingEverything,inf):-!.
setSize(abstractSet(AI,X),abstract(AI,SIZE)):-!,setSize(X,SIZE).
setSize(X,SIZE):-setToList(X,LIST),length(LIST,SIZE).

insertListIntoSet([],A,A).
insertListIntoSet([H|T],IN,OUT):-insertIntoSet(H,IN,IN1),insertListIntoSet(T,IN1,OUT).

  




ruleWrapper(LHS,RHS1,RHS2):-abstRule(LHS,RHS1,RHS2),rule(LHS,RHS1,RHS2).
ruleWrapper(LHS,RHS):-abstRule(LHS,RHS),rule(LHS,RHS).


closureNew(reachable,NEW,LASTNEW,_):-
  member(OLD1,LASTNEW),
  (ruleWrapper(OLD1,NEW);ruleWrapper(OLD1,NEW1,NEW2),(NEW=NEW1;NEW=NEW2))
  .

closureNew(generating,NEW,LASTNEW,GENERATING):-
  member(OLD1,LASTNEW),
  (((ruleWrapper(NEW,OLD1,OLD2);ruleWrapper(NEW,OLD2,OLD1)),inSet(OLD2,GENERATING))
   ;
   ruleWrapper(NEW,OLD1)
  )
.


incrementTransitionCounter:-
  nb_getval(noOfTransitions,NO_OF_TRANS),
  NO_OF_TRANS_INC is NO_OF_TRANS+1,
  (0 is NO_OF_TRANS_INC mod 1000 -> writeln('transitions':NO_OF_TRANS_INC);true),
  nb_setval(noOfTransitions,NO_OF_TRANS_INC).



closureNew(STEPTYPE,NEW,LASTNEW,SUPERSET,RESULTSET):-
  closureNew(STEPTYPE,NEW,LASTNEW,RESULTSET),
  notARegFragment(NEW),
  (printQueueSize->incrementTransitionCounter;true),
  inSet(NEW,SUPERSET),
  \+inSet(NEW,RESULTSET).


writeSet(X):-assoc_to_keys(X,L),writeList(L).


closureIt(STEPTYPE,SUPERSET,LASTNEWS,RESULTIN,RESULTOUT):-
  %writeList(LASTNEWS),
  insertSortedListIntoSet(LASTNEWS,RESULTIN,RESULTIT),
  nb_setval(noOfTransitions,0),
  notFailingSetOf(NEW,closureNew(STEPTYPE,NEW,LASTNEWS,SUPERSET,RESULTIT),NEWS),
  (printQueueSize->length(NEWS,NEWSL),write(queue_size=NEWSL),nb_getval(noOfTransitions,NO_OF_TRANS),writeln(', no of transitions':NO_OF_TRANS)
    ;
  true),
  (NEWS=[]
     ->RESULTIT=RESULTOUT%, writeln('phase SET:'),writeSet(RESULTOUT),writeln('phaseset end')
     ;closureIt(STEPTYPE,SUPERSET,NEWS,RESULTIT,RESULTOUT))
  .



getGenerating(TERMS,REACHABLE,GENERATING):-
  emptySet(generatingSet,EMPT),
  closureIt(generating,REACHABLE,[eps|TERMS],EMPT,GENERATING).

getReachable(IN,GENERATING,REACHABLE):-
  emptySet(reachableSet,EMPT),
  closureIt(reachable,GENERATING,[IN],EMPT,REACHABLE).

