%:-module(main, [ndmpctxswSMT/1,ctxbounding/2,testndmpctxsw/0]).

%:-public(ndmpctxswSMT/1,ctxbounding/2).

/** <main> Program entry points 
    
    @author Tomas Poch
*/


:-public(ndmpctxsw/2).
:-public(ctxbounding/3).

%% ndmpctxsw(+METHOD, +FORMFILE).
%
%  Load the pattern, threads and optionally asssertthread from the predicate database,
%  generate the formula and either solve it or write it into the file to be solved by
%  an SMT solver
%
%  Threads are stored in the predicate threads(THREADLIST)
%
%  Pattern is stored in the predicate pattern(P)
%
%  @param FORMFILE file name used to write the formula 
%  @param METHOD method used to solve the formula (either clpfd or yices) 
%  


ndmpctxsw(METHOD,FORMFILE):-
  pattern(P),
  threads(T),
  ndmpctxswForm(T,P,FORM,TERMS),
  solve(METHOD,TERMS,FORM,FORMFILE)
  .
  


%% ctxbounding(+METHOD,+K,+FORMFILE).
%
%  Load the threads and optionally asssertthread from the predicate database, generate
%  the pattern capturing all sequences of K context switches and generate the formula.
%
%  @param K number of context switches to consider 
%  @param FORMFILE file name used to write the formula 
%  @param METHOD method used to solve the formula (either clpfd or yices) 
%  
%
ctxbounding(METHOD,K,FORMFILE):-
  threads(T),
  generateCtxBoundingPattern(T,K,P),
  ndmpctxswForm(T,P,FORM,TERMS),
  solve(METHOD,TERMS,FORM,FORMFILE)
  .


%% solve(+METHOD,+TERMS,+FORM,+FORMFILE)
%
%  Solve the formula using the given method.
%  
%  Solve the formula or write it into the file to be solved by an SMT solver
% 
%  @param METHOD method used to solve the formula (either clpfd or yices) 
%  @param TERMS list of terms used by the grammar. Used as names for terminal variables. 
%  @param FORM formula to be solved
%  @param FORMFILE file name used to write the formula 
%

solve(METHOD,TERMS,FORM,FORMFILE):-
  ((METHOD=clpfd)->
    solveByProlog(METHOD,FORM)
  ;
    solveBySolver(METHOD,TERMS,FORM,FORMFILE)   
  ).


solveByProlog(METHOD,FORM):-
  solve(METHOD,FORM,RESULT),
  writeln(('PKImage'=RESULT)).

solveBySolver(METHOD,TERMS,FORM,FORMFILE):-
  open(FORMFILE,write,STREAM),
  generateSMT(METHOD,STREAM,TERMS,FORM),
  close(STREAM)
  .


%%      ndmpctxsw_chain(+THREAD, +TIDS, +PATTERNGRM, +TERMS, ?TVARS, -FORMS) is det.
%
%       Transform the given thread into formula. The thread consist from thread id, 
%       control-flow grammar and memory grammar. First, intersection of the control-flow
%       grammar with che memory grammar is created, then context switching rules are added
%       Finally, projection keeps only the yield/go terminals and the
%       ndmp_chain/5 predicate is called to transform the result grammar into formulas.
%
%       @param THREAD thread in form of thread(TID,CTRLFLOW,MEMGRM) term
%       @param TIDS list of thread identifiers (natural numbers)
%       @param PATTERNGRM pattern grammar
%       @param TERMS list of terminals produced by the result grammar 
%                    (ptrnWord(N) terms - one terminal for each word of the pattern)
%       @param TVARS list of prolog variables used to represent number of occurences of 
%                    the particular terminal from TERMS in the result formula 
%       @param FORMS result formula represented using clpfd operators and prolog variables
%
ndmpctxsw_chain(thread(TID,CTRLFLOWGRM,MEMGRM),TIDS,PATTERNGRM,TERMS,TVARS,FORMS):-
    generateEncodePcRules(CTRLFLOWGRM,ENCODERULES),
    intersectCfgWithReg(CTRLFLOWGRM,MEMGRM,cfGrm(AXIOMS,CTRLFLMEMRULES,SUPPORTRULES)),!,
    generateCtxRules(TID,TIDS,CTRLFLMEMRULES,CTXSWGR),!,
    append(CTRLFLMEMRULES,CTXSWGR,MULTITHREADEDGRM),!,
    projection(MULTITHREADEDGRM,[yield,go],RESULTRULES),!,
    append(ENCODERULES,SUPPORTRULES,SUPPORTRULES2),
    ndmp_chain(cfGrm(AXIOMS,RESULTRULES,SUPPORTRULES2),PATTERNGRM,TERMS,TVARS,FORMS).


emptyFormulaList(TVARS,form([],TVARS,[],[])).


%%      ndmpctxswForm_hlp(+THREADS, +TIDS, +PATTERNGRM, +TERMS, ?TVARS, -FORM) is det.
%%      ndmpctxswForm_hlp(+THREADS, +PATTERN, -FORM,-TERMS) is det.
%
%       Transform given threads into formula. Predicate invokes ndmpctxsw_chain/6 for each 
%       thread and ensures that the variables representing the same terminals in the formula are 
%       the same for each thread. The result formula is conjunction of formulas produced for 
%       each thread.
%
%       @param THREADS list of threads 
%       @param TIDS list of thread identifiers (natural numbers)
%       @param PATTERNGRM pattern grammar 
%       @param TERMS list of terminals produced by the result grammar 
%                    (ptrnWord(N) terms - one terminal for each word of the pattern)
%       @param TVARS list of prolog variables used to represent number of occurences of 
%                    the particular terminal from TERMS in the result formula 
%       @param FORMS result formula represented using clpfd operators and prolog variables
%       @param PATTERN pattern - list of words. Each word is a list of yield/go terminals
%
ndmpctxswForm_hlp([],_,PATTERNGRM,TERMS,TVARS,FORMS):-
  %emptyFormulaList(TVARS,FORMS).
  generateYieldGoGrammar(PATTERNGRM,YIELDGOGRM),
  ndmp_chain(YIELDGOGRM,PATTERNGRM,TERMS,TVARS,FORMS).

ndmpctxswForm_hlp([H|T],TIDS,PATTERNGRM,TERMS,TVARS,FORMS):-
  ndmpctxsw_chain(H,TIDS,PATTERNGRM,TERMS,TVARS,FORMS1),
  ndmpctxswForm_hlp(T,TIDS,PATTERNGRM,TERMS,TVARS,FORMS2),
  mergeFormulas(FORMS1,FORMS2,FORMS).

ndmpctxswForm_hlp(THREADS,PATTERN,FORM,TERMS):-
  get_time(TS1),
  patternToRules(PATTERN,PATTERNGRM,WCOUNT),
  ndmpterms(0,WCOUNT,TERMS,TVARS),!,
  collectTIDS(THREADS,TIDS),!,
  ndmpctxswForm_hlp(THREADS,TIDS,PATTERNGRM,TERMS,TVARS,FORM),
  get_time(TS2),
  DT is TS2-TS1,
  (printTime->write(','),writeln(totaltime=DT);true).



%%      ndmpctxswForm(+THREADS, +PATTERN, -FORM, -TERMS) is det.
%
%       Create the formula representing the concurrent computation of the given threads following the pattern. 
%
%       The solution of the formula represents the particular computation in terms of context swichtes. The computation
%       represented as the content of the shared memory at each switch admissible by all threads running from their initial
%       points to the terminal points. If there is no solution, one of the threads is deadlocked. 
%
%       @param THREADS list of threads
%       @param PATTERN pattern - list of words. Each word is a list of yield/go terminals
%       @param FORM generated formula 
%       @param TERMS list of terminals produced by the result grammar 
%
ndmpctxswForm(THREADSIN,PATTERN,FORM,TERMS):-
  treatNormalThreads(THREADSIN,THREADS),!, 
  ndmpctxswForm_hlp(THREADS,PATTERN,FORM,TERMS).


%% ndmpterms(+IDX,+PTRNSIZE, -TERMS, -TVARS)
%
%  Generate list of terminals and their respective formula vars that can occur in the grammar
%  after intersecting with the pattern and applying the homomorphism.
%  
%  @param IDX variable driving the recursion. Use 0 to invoke the predicate
%  @param PTRNSIZE size of the pattern
%  @param TERMS generated list of terminals [ptrnWord(0)..ptrnWord(PTRNSIZE-1)]
%  @param TVARS newly created prolog variables used to count the number of resp. terminals in the formula 
ndmpterms(IDX,IDX,[],[]).
ndmpterms(IDX,WCOUNT,[term(ptrnWord(IDX))|TERMTAIL],[_|VARTAIL]):-
  IDX<WCOUNT,IDX2 is IDX+1,
  ndmpterms(IDX2,WCOUNT,TERMTAIL,VARTAIL).


%%      ndmp_chain(+SYMCFGRM, +PATTERNGRM, +TERMS, ?TVARS, -FORMS) is det.
%
%       Intersect the given context-free grammar with the regular grammar, apply the
%       homomorphism that transforms the pattern words into single symbol and generate
%       the formula characterizing the Parikh image of the result.
%
%       @param SYMCFGRM context-free grammar 
%       @param PATTERNGRM regular grammar representing the pattern 
%       @param TERMS list of terminals produced by the result grammar 
%                    (ptrnWord(N) terms - one terminal for each word of the pattern)
%       @param TVARS list of prolog variables used to represent number of occurences of 
%                    the particular terminal from TERMS in the result formula 
%       @param FORMS result formula represented using clpfd operators and prolog variables
%
ndmp_chain(SYMCFGRM,PATTERNGRM,TERMS,TVARS,FORMS):-
    unsetRuleGrounding,
    %simulateGrammar(alllevels,SYMCFGRM),
    intersectCfgWithReg(SYMCFGRM,PATTERNGRM,cfGrm(PIAXIOM,PIRULES,SUPPORTRULES)),!,
    %simulateGrammar(alllevels,cfGrm(PIAXIOM,PIRULES,SUPPORTRULES)),
    patternHomomorph(PIRULES,HOMOMORPHRULES),!,
    sortClauseBodies(HOMOMORPHRULES,SYMRULES),!,
    (execute->simulateGrammar(alllevels,cfGrm(PIAXIOM,SYMRULES,SUPPORTRULES));true),
    instantiateRules(cfGrm(PIAXIOM,SYMRULES,SUPPORTRULES),TERMS,ACTIVENTERMS,INSTRULES),!,
    ndpk_chain(cfGrm(PIAXIOM,INSTRULES),TERMS,ACTIVENTERMS,TVARS,FORMS).


%%      ndmp_chain(+SYMCFGRMS, +PATTERNGRM, +TERMS, ?TVARS, -FORMS) is det.
%
%       Intersect each of the given context-free grammars with the regular grammar, apply the
%       homomorphism that transforms the pattern words into single symbol and generate
%       the formula characterizing the Parikh image of the intersection of all grammars.
%       For each grammar invokes ndmp_chain/5
%
%       @param SYMCFGRMS list of context-free grammars
%       @param PATTERNGRM regular grammar representing the pattern
%       @param TERMS list of terminals produced by the result grammar 
%                    (ptrnWord(N) terms - one terminal for each word of the pattern)
%       @param TVARS list of prolog variables used to represent number of occurences of 
%                    the particular terminal from TERMS in the result formula 
%       @param FORMS result formula represented using clpfd operators and prolog variables
%
ndmp([],_,_,TVARS,FORMS):-emptyFormulaList(TVARS,FORMS).
ndmp([H|T],PATTERNGRM,TERMS,TVARS,FORMS):-
  ndmp_chain(H,PATTERNGRM,TERMS,TVARS,FORMS1),
  ndmp(T,PATTERNGRM,TERMS,TVARS,FORMS2),
  mergeFormulas(FORMS1,FORMS2,FORMS).


%%      ndmp_chain(+SYMCFGRMS, +PATTERN, -RESULT) is det.
%
%       Intersect each of the given context-free grammars with the pattern, apply the
%       homomorphism that transforms the pattern words into single symbol, generate
%       the formula characterizing the Parikh image of the intersection of all grammars
%       and solve it using clpfd package.
%
%       @param SYMCFGRMS list of context-free grammars
%       @param PATTERN pattern
%       @param RESULT Parikh image (i-th word is the number of occurences of the i-th pattern word)
%
ndmp(SYMCFGRMS,PATTERN):-
  patternToRules(PATTERN,PATTERNGRM,WCOUNT),
  ndmpterms(0,WCOUNT,TERMS,TVARS),
  ndmp(SYMCFGRMS,PATTERNGRM,TERMS,TVARS,FORMS),
  solve(yices,TERMS,FORMS,'tmpfile.ys').


%%      ndpk_chain(+CFGRM,+TERMS,+NTERMS,?TVARS,-FORM)
%
%       Generate the formula characterizing the Parikh image of the given context free grammar
%
%       @param CFGRM context-free grammar
%       @param TERMS list of terminals produced by the grammar 
%                    (ptrnWord(N) terms - one terminal for each word of the pattern)
%       @param NTERMS list of nonterminals used by the grammar 
%       @param TVARS list of prolog variables used to represent number of occurences of 
%                    the particular terminal from TERMS in the result formula 
%       @param FORM result formula represented using clpfd operators and prolog variables
%

ndpk_chain(cfGrm(AXIOM,RULES),TERMS,NTERMS,TVARS,form(EQLIST,TVARS,RULEVARS,[REACHMAP])):-
    generateEquation(AXIOM,TERMS,NTERMS,TVARS,RULES,EQLIST,RULEVARS,REACHMAP).

isYield(term(yield(_))).

mapSingleTermRule(NT,T,rule(NT,T)).

generateYieldGoGrammar(regGrm(_,REGRULES,_,_),cfGrm(b,[rule(b,g,ax),rule(b,eps),rule(ax,y,b),rule(ax,eps)|RULES])):-
%generateYieldGoGrammar(regGrm(_,REGRULES,_,_),cfGrm(b,[rule(ax,g,ax),rule(ax,y,b),rule(ax,eps),rule(b,g,ax),rule(b,eps)|RULES])):-
    getTerminalsAndNonterminals(REGRULES,TERMS,_),
    partition(isYield,TERMS,YIELDTERMS,OTHERTERMS),
    maplist(mapSingleTermRule(g),OTHERTERMS,GORULES),
    maplist(mapSingleTermRule(y),YIELDTERMS,YIELDRULES),
    append(GORULES,YIELDRULES,RULES).
    
testndmpctxsw:-ndmpctxsw(yices,'input.ys').

setupTest:-consult('C:\\projects\\vanocka\\trunk\\examples\\assertMin_2threads'),consult('setImpl_assoc'),setUseAbstraction(local).
runtest:-ndmpctxsw(yices,'out.yices').
