/** <memgrm> Generate the regular grammar representing memory from the memory definition
@author Tomas Poch
*/
%signsucc(X,Y):-succ(X,Y).

%:-module(memgrm, [generateGmemDomain/2,generateMemGrm/3,inDomain/2,inDomainAbstract/2,signsucc/2]).

%% signsucc(+X,-Y).
%% signsucc(-X,+Y).
%% signsucc(+X,+Y).
%
%  succ predicate supporting negative numbers 
%
signsucc(X,Y):-var(Y),integer(X),!,Y is X+1.
signsucc(X,Y):-var(X),integer(Y),!,X is Y-1.
signsucc(X,Y):-integer(X),integer(Y),!,Y is X+1.
signsucc(_,_):-%trace,
  fail.


%%  generateRulesForVar(+TYPE,+NAME,+ASSERTMODE,-RULES)
%
%  Generate set of rules for one variable of the given type
%
%  @param TYPE type of variable. Following types are supported
%  * mutex - values are in {locked,unlocked} operations in {lock,unlock}. 
%            Used to implement mutual exclusion 
%  * bool - values in {true,false}, operations include assignment and comparison with a constant
%  * int(MIN,MAX)  - values in range MIN..MAX of integer numbers. Supports assignment, comparison, with a constant, increment, decrement
%  @param NAME name of the variable
%  @param ASSERTMODE either 'noerrorstate' or 'includeerrorstate'. In the former case, the assert action does not do anything while in the
%  latter case if the condition is not met, the variable is changed to the errmem value.
%  @param RULES result rules
%
%

generateRulesForVar(mutex,NAME,_,RULES):-
  RULES= [
    rule(NAME,unlocked,term(lock(NAME)),locked),
    rule(NAME,locked,term(unlock(NAME)),unlocked),
    rule(NAME,unlocked,term(unlock(NAME)),unlocked)
  ].

generateRulesForVar(bool,NAME,ERRORSTATE,RULES):-
  generateAsserts(bool,NAME,ERRORSTATE,ASSERTRULES),
  RULES = [
    (rule(NAME,top,term(_),top)),
    rule(NAME,V1,term(NAME==V1),V1),   
    (rule(NAME,V2,term(not(NAME==V3)),V2):-inDomain(bool,V2),V2\=V3),
    (rule(NAME,V5,term(NAME=V4),V4):-inDomain(bool,V5))
  |ASSERTRULES].

generateRulesForVar(int,NAME,ERRORSTATE,RULES):-
  generateRulesForVar(int(-1,5),NAME,ERRORSTATE,RULES).

generateRulesForVar(int(MAX),NAME,ERRORSTATE,RULES):-
  MIN is 0-MAX,
  generateRulesForVar(int(MIN,MAX),NAME,ERRORSTATE,RULES).

generateRulesForVar(int(MIN,MAX),NAME,ERRORSTATE,RULES):-
  generateAsserts(int(MIN,MAX),NAME,ERRORSTATE,ASSERTRULES),
  RULES = [
    (rule(NAME,top,term(_),top)),
    rule(NAME,V1,term(NAME==V1),V1),   
    (rule(NAME,V2,term(not(NAME==V3)),V2):-inDomain(int(MIN,MAX),V2),V2\=V3),
    (rule(NAME,V12,term(NAME=V4),V4):-inDomain(int(MIN,MAX),V12)),
    (rule(NAME,V1,term(inc(NAME)),V2):-signsucc(V1,V2)),
    (rule(NAME,V3,term(dec(NAME)),V4):-signsucc(V4,V3)),
    (rule(NAME,V5,term(NAME<K),V5):-V5\=top->V5<K;true),
    (rule(NAME,V5,term(NAME>K),V5):-V5\=top->V5>K;true),
    (rule(NAME,V6,term(NAME>=K),V6):-V6\=top->V6>=K;true),
    (rule(NAME,V6,term(NAME=<K),V6):-V6\=top->V6=<K;true)
  |ASSERTRULES].

generateRulesForVar(uint(MAX),NAME,ERRORSTATE,RULES):-
  generateRulesForVar(int(0,MAX),NAME,ERRORSTATE,RULES).

generateRulesForVar(abstract,NAME,_,RULES):-
  RULES = [
   rule(NAME,top,term(_),top)
  ].


%% only assert(false) is supported
generateAsserts(_,_,_,[]).
%%  generateAsserts(+TYPE,+NAME,+ASSERTMODE,-RULES)
%
%  Generate rules for the assert operation for the given variable. The assert operation supports only comparison with a constant.
%
%  @param TYPE type of variable. 
%  @param NAME name of the variable
%  @param ASSERTMODE either 'noerrorstate' or 'includeerrorstate'. 
%         In the former case, the assert action does not do anything while in the
%         latter case if the condition is not met, the variable is changed to the errmem value.
%  @param RULES result rules
%
%
%generateAsserts(bool,NAME,noerrorstate,RULES):-!,
%  RULES = [
%    rule(NAME,V4,term(assert(NAME==_)),V4) 
%  ]. 
%
%generateAsserts(_,NAME,noerrorstate,RULES):-!,
%  RULES = [
%    rule(NAME,V4,term(assert(NAME==_)),V4), 
%    rule(NAME,V5,term(assert(NAME<_)),V5),
%    rule(NAME,V6,term(assert(NAME>_)),V6),
%    rule(NAME,V7,term(assert(NAME=<_)),V7), 
%    rule(NAME,V8,term(assert(NAME>=_)),V8)
%  ].
%
%
%generateAsserts(bool,NAME,includeerrorstate,RULES):-!,
%  RULES = [
%    rule(NAME,V4,term(assert(NAME==V4)),V4),   
%    (rule(NAME,V5,term(assert(NAME==V6)),errmem):-inDomain(bool,V5),V5\=V6)
%  ]. 
% 
%
%generateAsserts(TYPE,NAME,includeerrorstate,RULES):-
%  RULES = [
%    rule(NAME,V4,term(assert(NAME==V4)),V4),   
%    (rule(NAME,V7,term(assert(NAME<K)),V7):-inDomain(TYPE,V7),V7<K),
%    (rule(NAME,V8,term(assert(NAME>K)),V8):-inDomain(TYPE,V8),V8>K),
%    (rule(NAME,V9,term(assert(NAME>=K)),V9):-inDomain(TYPE,V9),V9>=K),
%    (rule(NAME,V10,term(assert(NAME=<K)),V10):-inDomain(TYPE,V10),V10=<K),
%
%    (rule(NAME,V5,term(assert(NAME==V6)),errmem):-inDomain(TYPE,V5),V5\=V6),
%    (rule(NAME,V11,term(assert(NAME<K)),errmem):-inDomain(TYPE,V11),V11>=K),
%    (rule(NAME,V12,term(assert(NAME>K)),errmem):-inDomain(TYPE,V12),V12=<K),
%    (rule(NAME,V13,term(assert(NAME>=K)),errmem):-inDomain(TYPE,V13),V13<K),
%    (rule(NAME,V14,term(assert(NAME=<K)),errmem):-inDomain(TYPE,V14),V14>K)
%  ]. 

%% inDomain(+TYPE,-VALUE).
%% inDomain(+TYPE,+VALUE).
%
%  For the given type get one of its values. 
%
%  @param TYPE variable type 
%  @param VALUE value of the given type
%

%inDomain(_,VAL):-atom(VAL),(VAL=err).
inDomain(bool,false).
inDomain(bool,true).

inDomain(mutex,locked).
inDomain(mutex,unlocked).

inDomain(abstract,top).

inDomain(int(MIN,MAX),VAL):-(var(VAL);integer(VAL)),between(MIN,MAX,VAL).
inDomain(INT,VAL):-normalizeIntDef(INT,NORMINT),inDomain(NORMINT,VAL).

normalizeIntDef(int,int(-1,5)).
normalizeIntDef(int(MAX),int(MIN,MAX)):-MIN is 0-MAX.
normalizeIntDef(uint(MAX),int(0,MAX)).

code(bool,false,0).
code(bool,true,1).
code(mutex,locked,0).
code(mutex,unlocked,1).
code(int(MIN,_),VAL,CODE):-CODE is VAL-MIN.
code(INT,VAL,CODE):-normalizeIntDef(INT,NORMINT),code(NORMINT,VAL,CODE).
code(_,top,0).




%% initValue(+TYPE,-VALUE)
%
%  For the given type get the default value. 
%
%  @param TYPE variable type 
%  @param VALUE 
%
initValue(bool, false,2).
initValue(mutex, unlocked,2).
initValue(INT,INITVAL,CODELIMIT):-normalizeIntDef(INT,NORMINT),initValue(NORMINT,INITVAL,CODELIMIT).
initValue(int(MIN,MAX),0,CODELIMIT):-CODELIMIT is (MAX-MIN)+1.
initValue(abstract,top,1).

%% generateNonTerminal(+MEMDEFSIZE,-NTERM)
%
%  Generate term representing a non-terminal of the memory grammar. It takes the form s(g(..),l(..)) where
%  the arguments of inner terms (g and l) are prolog variables. The number of arguments correpsond to the
%  number of variables in the memory definition
%
%  @param MEMDEFSIZE term memDefSize(NG,NL), where NG is the number of global variables and NL is the number of local variables
generateNonTerminal(memDefSize(NG,NL),s(GLOB,LOC)):-functor(GLOB,g,NG),functor(LOC,l,NL).

unifyAllButI_hlp(IDX,IDX,[A1|B],[A2|B],A1,A2).
unifyAllButI_hlp(IDX,IDX,[A1|B],[A2|B],A1,A2).
unifyAllButI_hlp(IDX,I,[A|B1],[A|B2],VARFROM,VARTO):-I<IDX,succ(I,I2),unifyAllButI_hlp(IDX,I2,B1,B2,VARFROM,VARTO).
unifyAllButI_hlp(IDX,A,B,VARFROM,VARTO):-unifyAllButI_hlp(IDX,0,A,B,VARFROM,VARTO).

%% unifyAllButI(+POS,?NT1,?NT2,-VARFROM,-VARTO)
%
%  Given the pair of memory grammar non-terminals, unify all corresponding variables but the one specified by POS.
%  Used when putting grammars for distinct variables together to form the memory grammar. In particular, most operations
%  of the memory grammar modifies only one variable - the one that is not unified. The other remains unchanged.
%  
%  @param POS specifies the variable position whithin the non-terminal. 
%             Takes the form pos(MEMTYPE,IDX), where MEMTYPE, being either 'glob' or 'loc', select the global or local memory and IDX identifies the particular variable within the g/l term
%
%  @param NT1 memory grammar non-terminal
%  @param NT2 memory grammar non-terminal
%  @param VARFROM the variable from NT1 specified by POS
%  @param VARTO the variable from NT2 specified by POS
%
%  Example : unifyAllButI(pos(glob,1),s(g(G1,G2,G3),l(L1,L2,L3)),s(g(G4,G5,G6),l(L4,L5,L6)),VARFROM,VARTO) --> 
%            unifyAllButI(pos(glob,1),s(g(G1,G2,G3),l(L1,L2,L3)),s(g(G1,G5,G3),l(L1,L2,L3)),G2,G5) 
%
unifyAllButI(pos(glob,IDX),s(GLOB1,LOC),s(GLOB2,LOC),VARFROM,VARTO):-
  GLOB1=..[g|GLOB1L],
  GLOB2=..[g|GLOB2L],
  unifyAllButI_hlp(IDX,GLOB1L,GLOB2L,VARFROM,VARTO).

unifyAllButI(pos(loc,IDX),s(GLOB,LOC1),s(GLOB,LOC2),VARFROM,VARTO):-
  LOC1=..[l|LOC1L],
  LOC2=..[l|LOC2L],
  unifyAllButI_hlp(IDX,LOC1L,LOC2L,VARFROM,VARTO).

%% unifyAllButName(+NAME,+MEMDEF,?NT1,?NT2,-VARFROM,-VARTO)
%
%  Given the pair of memory grammar non-terminals, unify all corresponding variables but the one specified by NAME.
%  Used when putting grammars for distinct variables together to form the memory grammar. In particular, most operations
%  of the memory grammar modifies only one variable - the one that is not unified. The other remains unchanged.
%
%  @param NAME   variable name
%  @param MEMDEF memory definition. Must contain the variable of the given name
%  @param NT1 memory grammar non-terminal
%  @param NT2 memory grammar non-terminal
%  @param VARFROM the variable from NT1 specified by POS
%  @param VARTO the variable from NT2 specified by POS
unifyAllButName(NAME,MEMDEF,S1,S2,VARFROM,VARTO):-
  getPositionByName(NAME,MEMDEF,POS),unifyAllButI(POS,S1,S2,VARFROM,VARTO).

%% getPositionByName(+NAME,+MEMDEF,-POS)
%
%  Get the position for the given name.
%
%  @param NAME   variable name
%  @param MEMDEF memory definition. Must contain the variable of the given name
%  @param POS position of the variable within the memory grammar non-terminals
getPositionByName(NAME,memDef(GLOBS,_),pos(GORL,IDX)):-
  (nth0(IDX,GLOBS,_-NAME);nth0(IDX,GLOBS,_-NAME=_)),GORL=glob.

getPositionByName(NAME,memDef(_,LOCS),pos(GORL,IDX)):-
  (nth0(IDX,LOCS,_-NAME);nth0(IDX,LOCS,_-NAME=_)),GORL=loc.


%% getTypeByName(+NAME,+MEMDEF,-TYPE)
%
%  Get the type of the given variable
%
%  @param NAME variable name  
%  @param MEMDEF memory definition. Must contain the variable of the given name
%  @param TYPE type of the variable
%
getTypeByName(NAME,memDef(GLOB,LOCS),TYPE):-
  (nth0(IDX,LOCS,TYPE-NAME);
  nth0(IDX,LOCS,TYPE-NAME=_);
  nth0(IDX,GLOB,TYPE-NAME);
  nth0(IDX,GLOB,TYPE-NAME=_)),!.

getVarByPosition_hlp(IDX,GORLTERM,VAR):-
  GORLTERM=..[_|LST],
  nth0(IDX,LST,VAR).

%% getVarByPosition(+POS,?NTERM,-VAR)
%
% Get the variable from memory grammar non-terminal specified by POS
%
%
%  @param POS position of the variable within the memory grammar non-terminals
%  @param NTERM memory grammar non-terminal
%  @param VAR variable from the NTERM specified by POS
%
getVarByPosition(pos(glob,IDX),s(GLOB,_),VAR):-getVarByPosition_hlp(IDX,GLOB,VAR).
getVarByPosition(pos(loc,IDX),s(_,LOC),VAR):-getVarByPosition_hlp(IDX,LOC,VAR).

%% getVarByPosition(+NAME,+MEMDEF,?NTERM,-VAR)
%
% Get the variable from memory grammar non-terminal specified by the name
%
%  @param NAME   variable name
%  @param MEMDEF memory definition. Must contain the variable of the given name
%  @param NTERM memory grammar non-terminal
%  @param VAR variable from the NTERM specified by POS
%
getVarByName(NAME,MEMDEF,S,VAR):-
  getPositionByName(NAME,MEMDEF,POS),
  getVarByPosition(POS,S,VAR).

%this is ugly. Rewrite  
multiplyRule(noerrorstate,RULEIN,transitionVars(IN,OUT,VARIN,VAROUT),MULTRULE):-
  ((RULEIN=rule(_,VARIN,term(TERM),VAROUT),BODY=true);RULEIN=(rule(_,VARIN,term(TERM),VAROUT):-BODY)),
  (var(VAROUT);VAROUT\=errmem),MULTRULE=(rule(IN,term(TERM),OUT):-BODY,valid(gmem,OUT)).

multiplyRule(includeerrorstate,RULEIN,transitionVars(IN,OUT,VARIN,VAROUT),MULTRULE):-
  (((RULEIN=rule(_,VARIN,term(assert(A)),VAROUT),BODY=true);RULEIN=(rule(_,VARIN,term(assert(A)),VAROUT):-BODY))
  ,atom(VAROUT),VAROUT=errmem)
    ->
      MULTRULE=(rule(IN,term(assert(A)),errmem):-BODY)
    ;
      multiplyRule(noerrorstate,RULEIN,transitionVars(IN,OUT,VARIN,VAROUT),MULTRULE)
  .

%% multiplyRules(+ASSERTMODE,+VARRULES,+POS,+MEMDEFSIZE,-MULTRULES).
%
% For the given rules for a single variable, generate rules relating the memory grammar non-terminals.
% The generated rule modifies just one variable and the other remain unchanged
%
%  @param ASSERTMODE either 'noerrorstate' or 'includeerrorstate'. In the latter case, 
%         once an variable is being changed to the errmem, change the whole memory to errmem. 
%  @param VARRULES list of rules for a single variable
%  @param POS position of the variable within memory grammar non-terminal
%  @param MEMDEFSIZE number of global and local variables
%  @param MULTRULES result rules
% 
multiplyRules(_,[],_,_,[]).
multiplyRules(ERRORSTATE,[RULE|VARRULES],POS,MEMDEFSIZE,[MULTRULE|MULTRULES]):-
  generateTransitionVars(POS,MEMDEFSIZE,TVARS),!,
  multiplyRule(ERRORSTATE,RULE,TVARS,MULTRULE),
  multiplyRules(ERRORSTATE,VARRULES,POS,MEMDEFSIZE,MULTRULES).

generateTransitionVars(POS,MEMDEFSIZE,transitionVars(IN,OUT,VARIN,VAROUT)):-
  generateNonTerminal(MEMDEFSIZE,IN),
  generateNonTerminal(MEMDEFSIZE,OUT),
  unifyAllButI(POS,IN,OUT,VARIN,VAROUT).
  
generateSingleVarTransitionRules(_,_,[],_,[]).
generateSingleVarTransitionRules(ERRORSTATE,pos(T,IDX),[VARDEF|VARDEFS],MEMDEFSIZE,RULES):-
  (VARDEF=(TYPE-NAME);VARDEF=(TYPE-NAME=_)),
  succ(IDX,IDX2),
  generateSingleVarTransitionRules(ERRORSTATE,pos(T,IDX2),VARDEFS,MEMDEFSIZE,RESTRULES),
  generateRulesForVar(TYPE,NAME,ERRORSTATE,VARRULES),
  multiplyRules(ERRORSTATE,VARRULES,pos(T,IDX),MEMDEFSIZE,MULTRULES),
  append(MULTRULES,RESTRULES,RULES).

getMemDefSize(memDef(GLOBALS,LOCALS),memDefSize(GSIZE,LSIZE)):-
  length(GLOBALS,GSIZE),
  length(LOCALS,LSIZE).

%% generateSingleVarTransitionRules(+ASSERTMODE,+MEMDEF,+MEMDEFSIZE,-MULTRULES).
%
% Generate rules of the memory grammar representing actions involving only single variable.
%
%  @param ASSERTMODE either 'noerrorstate' or 'includeerrorstate'
%  @param MEMDEF memory definition. 
%  @param MEMDEFSIZE number of global and local variables
%  @param MULTRULES result rules
% 
generateSingleVarTransitionRules(ERRORSTATE,memDef(GLOBALS,LOCALS),MEMDEFSIZE,RULES):-
  generateSingleVarTransitionRules(ERRORSTATE,pos(glob,0),GLOBALS,MEMDEFSIZE,GRULES),
  generateSingleVarTransitionRules(ERRORSTATE,pos(loc,0),LOCALS,MEMDEFSIZE,LRULES),
  append(GRULES,LRULES,RULES).
  
generateCrossOpTransitionRule(MEMDEF,MEMDEFSIZE,TARGET=SOURCE,T):-
  getPositionByName(TARGET,MEMDEF,POS),
  generateTransitionVars(POS,MEMDEFSIZE,transitionVars(IN,OUT,_,VAROUT)),!,
  getVarByName(SOURCE,MEMDEF,IN,SOURCEVAR),
  getTypeByName(TARGET,MEMDEF,TTYPE),
  getTypeByName(SOURCE,MEMDEF,STYPE),
  ((TTYPE\=abstract,STYPE\=abstract)->
     T=(rule(IN,term(TARGET=SOURCE),OUT):-VAROUT=SOURCEVAR,valid(gmem,OUT))
     ;
     T=(rule(IN,term(TARGET=SOURCE),OUT):-valid(gmem,OUT))
  )
  .

generateCrossOpTransitionRule(MEMDEF,MEMDEFSIZE,TARGET==SOURCE,T):-
  generateNonTerminal(MEMDEFSIZE,IN),
  getVarByName(SOURCE,MEMDEF,IN,SOURCEVAR),
  getVarByName(TARGET,MEMDEF,IN,TARGETVAR),
  getTypeByName(TARGET,MEMDEF,TTYPE),
  getTypeByName(SOURCE,MEMDEF,STYPE),
  ((TTYPE\=abstract,STYPE\=abstract)->
     T=(rule(IN,term(TARGET==SOURCE),IN):-TARGETVAR=SOURCEVAR,valid(gmem,IN))
     ;
     T=(rule(IN,term(TARGET==SOURCE),IN):-valid(gmem,IN))
  )
  .




%% generateCrossOpTransitionRules(+MEMDEF,+MEMDEFSIZE,+OPS,-RULES).
%
% Generate rules of the memory grammar representing actions involving more than one variable.
%
%  @param MEMDEF memory definition. 
%  @param MEMDEFSIZE number of global and local variables
%  @param OPS list of operations. Currently only assignment is supported (TARGET=SOURCE)
%  @param RULES result rules
% 
generateCrossOpTransitionRules(_,_,[],[]).
generateCrossOpTransitionRules(MEMDEF,MEMDEFSIZE,[OP|OPS],[T|TS]):-
  generateCrossOpTransitionRule(MEMDEF,MEMDEFSIZE,OP,T),
  generateCrossOpTransitionRules(MEMDEF,MEMDEFSIZE,OPS,TS).

generateValidRule(_,_,[],[],IN,IN).

generateValidRule(AI,IDX,[VARDEF|VARDEFS],[VAR|VARS],IN,BODY):-
  (IDX<AI->
    generateValidRuleForVar(VARDEF,VAR,FORTHISVAR),
    BODY=(FORTHISVAR,BODYREST)
  ;
    VAR=top,
    BODY=BODYREST
  ),
  succ(IDX,IDX2),
  generateValidRule(AI,IDX2,VARDEFS,VARS,IN,BODYREST).

generateValidRuleForVar(TYPE-_=_,VAR,RULE):-!,generateValidRuleForVar(TYPE,VAR,RULE).
generateValidRuleForVar(TYPE-_,VAR,RULE):-!,generateValidRuleForVar(TYPE,VAR,RULE).
generateValidRuleForVar(TYPE,VAR,inDomain(TYPE,VAR)).

%% generateValidRule(+MEMDEF,-VALIDRULE)
%
% Generate predicate valid(gmem,NT) that grounds all variables in NT to one of valid values.
% 
%  @param MEMDEF memory definition. 
%  @param VALIDRULE valid rule
%
generateValidRule(memDef(GLOBALS,LOCALS),abstInfo(GAI,LAI),(valid(gmem,s(GLOBS,LOCS)):-BODY)):-
  generateValidRule(LAI,0,LOCALS,LOCVARS,true,LOCBODY),
  generateValidRule(GAI,0,GLOBALS,GLOBVARS,LOCBODY,BODY),
  GLOBS=..[g|GLOBVARS],
  LOCS=..[l|LOCVARS].
  

generateValidRuleMap(MEMDEF,RULEMAP):-
  getMaxAbstractInfo(MEMDEF,MAXABSTINFO),
  generateValidRuleMap(MEMDEF,abstInfo(0,0),MAXABSTINFO,RULEMAP).

generateValidRuleMap(MEMDEF,ABSTINFO,MAXABSTINFO,[validRuleMap(ABSTINFO,RULE)|RULEMAP]):-
  generateValidRule(MEMDEF,ABSTINFO,RULE),
  (getNextAbstractInfo(ABSTINFO,MAXABSTINFO,NEXTABSTINFO)->
    generateValidRuleMap(MEMDEF,NEXTABSTINFO,MAXABSTINFO,RULEMAP)
  ;
    RULEMAP=[validRuleMap(noAbstAtAll,RULE)]
  ).


generateInitState_hlp([],[]).
generateInitState_hlp([_-_=INITVAL|VARDEFS],[INITVAL|VALUES]):-!,
  generateInitState_hlp(VARDEFS,VALUES).
generateInitState_hlp([TYPE-_|VARDEFS],[VAL|VALUES]):-
  generateInitState_hlp(VARDEFS,VALUES),
  initValue(TYPE,VAL,_).

%% generateInitState(+MEMDEF,-INITNT)
%
%  Generate the initial non-terminal of memory grammar. Assigns the default values to all variables
%
%  @param MEMDEF memory definition. 
%  @param INITNT initial non-terminal 
generateInitState(memDef(GLOBALS,LOCALS),s(GLOBS,LOCS)):-
  generateInitState_hlp(GLOBALS,GLIST),
  generateInitState_hlp(LOCALS,LLIST),
  GLOBS=..[g|GLIST],
  LOCS=..[l|LLIST].


generateMemEncodeRule_hlp([],[],1,0,true).
generateMemEncodeRule_hlp([T|TTAIL],[VAR|VARTAIL],MAXDOM,EXPR,FLA):-
  generateMemEncodeRule_hlp(TTAIL,VARTAIL,MAXDOM2,EXPR2,FLA2),
  (T=(TYPE-_=_);T=(TYPE-_)),
  initValue(TYPE,_,MAXCODE),
  FLA=(FLA2,code(TYPE,VAR,VARCODE)),
  EXPR=( EXPR2*MAXCODE+VARCODE),
  MAXDOM is MAXDOM2 * MAXCODE.

generateMemEncodeInfo(memDef(GLOBALS,LOCALS),encodeInfo([RULE,encodeNT(gmem,errmem,MAXRNG)],MAXRNG1)):-
  generateMemEncodeRule_hlp(GLOBALS,GVECT,GMAXRNG,GEXPR,GFLA),
  generateMemEncodeRule_hlp(LOCALS,LVECT,LMAXRNG,LEXPR,LFLA),
  GLOBS=..[g|GVECT],
  LOCS=..[l|LVECT],
  RULE=(encodeNT(gmem,s(GLOBS,LOCS),CODE):-GFLA,LFLA,CODE is LEXPR*GMAXRNG+GEXPR),
  MAXRNG is LMAXRNG*GMAXRNG,
  succ(MAXRNG,MAXRNG1).
  



%% generateMemGrm(ERRORSTATE,MEMDEF,MEMGRM).
%
%  Generate the memory grammar for the given memory definition
%
%  @param ASSERTMODE either 'noerrorstate' or 'includeerrorstate'. 
%         In the former case, the assert action does not do anything while in the
%         latter case if the condition is not met, the variable is changed to the errmem value.
%  @param MEMDEF memory definition. 
%

generateMemGrm(ERRORSTATE,memDef(GLOBALS,LOCALS),MEMGRM):-
%  append(GLOBALS,LOCALS,VARSDECL),
%  generateCrossOps(VARS,VARS,CROSSOPS),
  generateMemGrm(ERRORSTATE,memDef(GLOBALS,LOCALS,[]),MEMGRM).

%generateCrossOps([],_,[]).
%generateCrossOps([V|VARS1],VARS,CROSSOPS):-
%  generateCrossOps1(V,VARS,CROSSOPS1),
%  generateCrossOps(VARS1,VARS,CROSSOPS2),
%  append(CROSSOPS1,CROSSOPS2,CROSSOPS).
%
%generateCrossOps1(_,[],[]).
%generateCrossOps1(V1,[V2|VARS],[V1=V2,V1==V2|CROSSOPS1]):-
%  generateCrossOps1(V1,VARS,CROSSOPS1).

generateMemGrm(ERRORSTATE,memDef(GLOBALS,LOCALS,CROSSOPS),MEMGRM):-
  getMemDefSize(memDef(GLOBALS,LOCALS),MEMDEFSIZE),
  generateCrossOpTransitionRules(memDef(GLOBALS,LOCALS),MEMDEFSIZE,CROSSOPS,TRANSITIONRULES2),!,
  generateSingleVarTransitionRules(ERRORSTATE,memDef(GLOBALS,LOCALS),MEMDEFSIZE,TRANSITIONRULES1),!,
  append(TRANSITIONRULES1,TRANSITIONRULES2,TRANSITIONRULES),!,
  generateValidRuleMap(memDef(GLOBALS,LOCALS),VALIDRULEMAP),!,
  generateInitState(memDef(GLOBALS,LOCALS),INIT),!,
  generateMemEncodeInfo(memDef(GLOBALS,LOCALS),ENCODEINFO),
  append(VALIDRULEMAP,TRANSITIONRULES,RULES),  
  MEMGRM=regGrm(INIT,[%(rule(errmem,term(err),errmem)),
                                rule(errmem,eps),
				rule(s(_,_),eps),
				rule(s(_,_),term(assert(false)),errmem),
				(rule(s(G,L),term(and(X,Y)),s(G2,L2)):-rule(s(G,L),term(X),s(G1,L1)),rule(s(G1,L1),term(Y),s(G2,L2))),
				(rule(s(G,L),term(or(X,Y)),s(G,L)):-rule(s(G,L),term(X),s(G,L));rule(s(G,L),term(Y),s(G,L))),
				rule(s(G,L),term(assert(true)),s(G,L)),
				rule(s(G,L),term(true),s(G,L))
				,valid(gmem,errmem)
				|RULES],gmem,ENCODEINFO),
  %writeRegGrm(MEMGRM),
  (debug->writeRegGrm(MEMGRM);true)
  .

writeRegGrm(regGrm(INIT,RULES,TOKEN,_)):-
  write('Regular grammar, init='),write(INIT),write(' token='),writeln(TOKEN),
  writeln('rules:'),
  writeList(RULES),
  writeln('end of regural grammar'),nl.


%% generateGmemDomain(+MEMDEF,-NT)
%
%  Generate valid grounded non-terminal term 
%
%  @param MEMDEF memory definition. 
%  @param NT valid non-terminal term 
%  
generateGmemDomain(memDef(GLOBALS,A,_),X):-generateGmemDomain(memDef(GLOBALS,A),X).
generateGmemDomain(memDef(GLOBALS,_),X):-
  length(GLOBALS,GL),
  generateValidRule(GL,0,GLOBALS,GLOBVARS,true,BODY),
  G=..[g|GLOBVARS],
  notFailingSetOf(G,BODY,X). 
