%:-module(preprocessThreads, [collectTIDS/2,generateCtxBoundingPattern/3,treatAssertThread/2, treatNormalThreads/2]).

/** <preprocessThreads> Alter the control flow grammars to generate only the words representing assertion violation
    
    @author Tomas Poch
*/


%% getLHSFromRules(+RULES,-LHSLIST).
%
%  Collect the left-hand side non-terminals from the list of rules
%
%  @param RULES    list of grammar rules
%  @param LHSLIST  result set of non-terminals
getLHSFromRules(GRMRULES,LHSLIST):-
  empty_assoc(E),
  getLHSFromRules_assoc(GRMRULES,E,ASSOC),
  assoc_to_keys(ASSOC,LHSLIST).
  

getLHSFromRules_assoc([],E,E).
getLHSFromRules_assoc([R|RULES],ASSOC1,ASSOC3):-
  getLHSFromRules_assoc(RULES,ASSOC1,ASSOC2),
  getLHSFromRule(R,ASSOC2,ASSOC3).


getLHSFromRule(rule(A,_),ASSOC1,ASSOC):-
  put_assoc(A,ASSOC1,1,ASSOC).

getLHSFromRule(rule(A,_,_),ASSOC1,ASSOC):-
  put_assoc(A,ASSOC1,1,ASSOC).

generateValidPCRules([],[]).
generateValidPCRules([PC|PCS],[validpc(PC)|RULES]):-
  generateValidPCRules(PCS,RULES).


generateEncodePcRules_hlp(_,[],[]).
generateEncodePcRules_hlp(IDX,[ARG|ARGS],[R1,R2|RULES]):-
  R1=(encodeNT(ARG,IDX)),
  succ(IDX,IDX2),
  R2=(encodeNT(z(ARG),IDX2)),
  succ(IDX2,IDX3),
  generateEncodePcRules_hlp(IDX3,ARGS,RULES).



generateEncodePcRules(cfGrm(_,RULES,_),ENCODERULES1):-
  getLHSFromRules(RULES,LHSLIST),
  generateEncodePcRules_hlp(0,LHSLIST,ENCODERULES1).


%% generateErrorRules(+NTERMS,-ERRORRULES).
%
%  Generate error rule for each non-terminal. The error rule allows to 
%  terminate each derivation by generating the symbol term(err).
%
%  When intersected with memory grammar (producing term(err) in error states), 
%  each computation can be terminated immediately when an error state is reached
%  
%  @param +NTERMS list of terms
%  @param +ERRORRULES list of error rules
generateErrorRules([],[]).
generateErrorRules([NT|NTT],[rule(NT,term(err))|RT]):-
  generateErrorRules(NTT,RT).


%% collectTIDS(+THREADS,-TIDS)
%
%  Collect thread ids
%
%  @param THREADS list of threads
%  @param TIDS list of thread IDs seen in THREADS
% 
collectTIDS([],[]).
collectTIDS([T|TT],[TID|TIDT]):-
  T=thread(TID,_,_),
  collectTIDS(TT,TIDT).


  

%% treatNormalThread(+TIN,-TOUT)
%
% Preprocess the thread. From the memory definition generate the memory grammar.
% Add new rule rewriting the final nonterminal into empty symbol. This way is ensured
% that even an assert thread can be used in the role of common thread - the derivation
% is terminated when the thread reaches its final point. If there is no "final" nonterminal, 
% the rule is just harmlessly unreachable
%
% @param TIN input thread consisting of TID,control-flow grammar and either memory definition or memory grammar. In the former case the memory definition is transformed into the memory grammar
% @param TOUT preprocessed thread
%
treatNormalThread(thread(ID,CFGRMIN,MEMDEF),thread(ID,CFGRMOUT,MEMGRM)):-
  treatNormalCFGrm(CFGRMIN,CFGRMOUT),
  generateMemGrm(noerrorstate,MEMDEF,MEMGRM),
  CFGRMOUT=cfGrm(_,RULES,_),
  printDefinitionSizes(RULES,MEMDEF)
  .

treatNormalThread(thread(ID,CFGRMIN,regGrm(INIT,R,T)),thread(ID,CFGRMOUT,regGrm(INIT,R,T))):-
  treatNormalCFGrm(CFGRMIN,CFGRMOUT).

%% treatNormalCFGrm(+CFGIN,-CFOUT)
%
% Preprocess the control-flow grammar 
%
% If there are no support rules, generate support rules (validpc/2) defining
% the set of non-terminals in the control-flow grammar. It is used to ground the set
% of non-active non-terminal in the context switching rules
%
% Add new rule rewriting the final nonterminal into empty symbol. This way is ensured
% that even an assert thread can be used in the role of common thread - the derivation
% is terminated when the thread reaches its final point. If there is no "final" nonterminal, 
% the rule is just harmlessly unreachablet

% @param CFGIN input controlflow grammar 
% @param CFGOUT preprocessed grammar
treatNormalCFGrm(cfGrm(START,RIN),cfGrm(START,ROUT,S)):-
  getTerminalsAndNonterminals(RIN,_,NTERMSL),
  generateErrorRules(NTERMSL,ERRRULES),
  append(RIN,ERRRULES,ROUT),
  generateValidPCRules(NTERMSL,S).
  
treatNormalCFGrm(cfGrm(START,RIN,S),cfGrm(START,ROUT,S)):-
  getTerminalsAndNonterminals(RIN,_,NTERMSL),
  generateErrorRules(NTERMSL,ERRRULES),
  append(RIN,ERRRULES,ROUT).
  


%% treatNormalThreads(+TIN,-TOUT)
%
% Preprocess the list of threads by invoking treatNormalThread/2
%
% @param TIN list of input threads
% @param TOUT preprocessed threads
%
treatNormalThreads([],[]).
treatNormalThreads([H|T],[RH|TH]):-
  treatNormalThread(H,RH),
  treatNormalThreads(T,TH).

%% generateCtxBoundingPattern(+THREADS,+K,-PTRN)
% 
% Generate the pattern capturing all sequences of K context switches
%
% The pattern takes advantage that the context switching rules produce strictly
% a pair of yield/go terminals for each context switch
%
% Thus, the pattern looks like this
%
% ( go(0)*go(1)*...*go(TIDMAX)*yield(GMEM_1)*yield(GMEM_2)*...yield(GMEM_MAX)* )^k
%
% For each context switch the pattern enumerates all thread ids to express that
% whatever thread in the system can run (go fragment) and then enumerates all
% states of the global memory (yield fragment). The fragments are repeated K times.
%
% Apparently, the size of the pattern is exponential to the size of global memory definition
%
% @param THREADS threads used to collect thread IDs (for go fragment). The memory definition
% of the first thread is used to generate all states of global memory (yield fragment)
% @param K number of context switches
% @param PTRN the generated pattern
generateCtxBoundingPattern(THREADS,K,PTRN):-
  collectTIDS(THREADS,TIDS),
  THREADS=[thread(_,_,MEMDEF)|_],
  generateGmemDomain(MEMDEF,GMEMLIST),
  ctxBoundingOptionList(go,TIDS,TIDFRAGMENT),
  ctxBoundingOptionList(yield,GMEMLIST,GMEMFRAGMENT),
  concatFragments(TIDFRAGMENT,GMEMFRAGMENT,K,PTRN).

%% ctxBoundingOptionList(+TNAME,+OPTLIST,-FRAGMENT)
% 
% Generate the fragment of pattern.
%
% Given the term name and list of options generate the list of single item lists
% containing the term with single argument taken from the option list
%
% @param TNAME term name used in the fragment - either yield or go
% @param OPTLIST list of options to enumerate in the fragment
% @param FRAGMENT fragment of the pattern
%
ctxBoundingOptionList(_,[],[]).
ctxBoundingOptionList(TNAME,[H|T],[[ITEM]|TAIL]):-
  ITEM=..[TNAME,H],
  ctxBoundingOptionList(TNAME,T,TAIL).

%% concatFragments(+GOFRAG,+YIELDFRAG,+K,-RESULT)
%
% Copy the go/yield fragment pair K times and produce the result pattern
%
% @param GOFRAG go fragment
% @param YIELDFRAG yield fragment
% @param K number of context switches
% @param RESULT pattern
concatFragments(_,_,0,[]).
concatFragments(GOFRAG,YIELDFRAG,K,RESULT):-
  succ(K1,K),
  concatFragments(GOFRAG,YIELDFRAG,K1,RESULT1),
  append([GOFRAG,YIELDFRAG,RESULT1],RESULT).
 

printDefinitionSizes(CFGRMRULES,MEMDEF):-
  getTerminalsAndNonterminals(CFGRMRULES,_,NTERMSL),
  length(NTERMSL,NTERMSLS),writeln(program_locations:NTERMSLS),
  (MEMDEF=memDef(GLOBALS,LOCALS);MEMDEF=memDef(GLOBALS,LOCALS,_)),
  append(GLOBALS,LOCALS,VARS),
  getMemDefSpaceSize(VARS,VARSSIZE),
  TOTAL is 2* NTERMSLS * VARSSIZE*VARSSIZE,
  writeln(2* NTERMSLS * VARSSIZE*VARSSIZE = TOTAL).
 
getMemDefSpaceSize([],1).
getMemDefSpaceSize([VAR|VARS],RES):-
  getMemDefSpaceSize1(VAR,RES2),
  getMemDefSpaceSize(VARS,RES1),
  RES is RES1 * RES2.


getMemDefSpaceSize1(TYPE-NAME=_,RES):-
  getMemDefSpaceSize2(TYPE,NAME,RES).
  
getMemDefSpaceSize1(TYPE-NAME,RES):-
  getMemDefSpaceSize2(TYPE,NAME,RES).

getMemDefSpaceSize2(TYPE,NAME,RES):-
  notFailingSetOf(VAL,inDomain(TYPE,VAL),VALS),
  length(VALS,RES),
  writeln(TYPE:NAME=RES).


