/** <projection> Replace the given terminals in the grammar by silent action
@author Tomas Poch
*/

%projection(INGRM,ALPH,OUTGRM)

%:-module(projection, [patternHomomorph/2, projection/3]).

%% projection(+RULES,+ALPH,-PROJRULES)
%
%  Given the list of rules modify those producing a terminal which does not belong to ALPH to produce eps (silent action)
%
%  @param RULES list of input rules
%  @param ALPH  terminals to keep - functor name or list of functor names
%  @param PROJRULES list of the projection grammar rules 
%
projection([],_,[]).
projection([H|T],ALPH,[R|RT]):-
  proj_hlp(H,ALPH,R),
  projection(T,ALPH,RT).

doesBelongTo(_,[]):-fail.
doesBelongTo(A,[H|_]):-doesBelongTo1(A,H),!.
doesBelongTo(A,[_|T]):-doesBelongTo(A,T).
doesBelongTo1(A,A).
doesBelongTo1(A,ALPH):-functor(A,ALPH,_).

proj_hlp((RULE:-COND),ALPHTOKEEP,(RESULTRULE:-COND)):-
  proj_hlp(RULE,ALPHTOKEEP,RESULTRULE).

proj_hlp(rule(X,Y),_,rule(X,Y)):-Y\=term(_).
proj_hlp(rule(X,term(T)),ALPHTOKEEP,rule(X,term(T))):-doesBelongTo(T,ALPHTOKEEP).
proj_hlp(rule(X,term(T)),ALPHTOKEEP,rule(X,eps)):- \+doesBelongTo(T,ALPHTOKEEP).
proj_hlp(rule(X,term(T),Y),ALPHTOKEEP,rule(X,term(T),Y)):-doesBelongTo(T,ALPHTOKEEP).
proj_hlp(rule(X,term(T),Y),ALPHTOKEEP,rule(X,Y)):- \+doesBelongTo(T,ALPHTOKEEP).
proj_hlp(rule(X,Y,Z),_,rule(X,Y,Z)):-Y\=term(_).

%%  patternHomomorph(+RULES,-RESULT)
%
%  Modify the list of given rules to silent the symbols when moving within single pattern words.
%  When moving to the beginning of a Ith pattern word  terminal ptrnWord(I) is produced
%
%  Assumes that non-terminals used in RULES takes the form pr(X,p(I,J),p(K,L)) - result of intersection of CFG with a pattern
%
%  @param RULES list of input rules
%  @param RESULT list modified rules 
%
patternHomomorph([],[]).
patternHomomorph([H1|T1],RESULT):-
  homomorph_hlp(H1,R),!,
  patternHomomorph(T1,R2),
  append(R,R2,RESULT).




appendCondToRules(_,[],[]).
appendCondToRules(true,L,L).
appendCondToRules(COND,[H|T],[H2|T2]):-COND\=true,
  appendCondToRule(COND,H,H2),
  appendCondToRules(COND,T,T2).

appendCondToRule(COND,HEAD,(HEAD:-COND)):-HEAD\=(_:-_).
appendCondToRule(COND,(HEAD:-RULES),(HEAD:-RULES,COND)).

homomorph_hlp((HEAD:-COND),RESULTRULES):-
  homomorph1(HEAD,RESULTRULES1),
  appendCondToRules(COND,RESULTRULES1,RESULTRULES).

homomorph_hlp(HEAD,RESULT):-HEAD\=(_:-_),
  homomorph1(HEAD,RESULT).

homomorph1(rule(pr(X,p(I,J),K),term(_),pr(Y,p(I2,J2),K)),
                    [(rule(pr(X,p(I,J),K),pr(Y,p(I2,J2),K)):-J2\=0),
                     (rule(pr(X,p(I,J),K),term(ptrnWord(PW)),pr(Y,p(I2,J2),K)):-J2=0,(I=I2,PW=I;J=0,PW=I2))]
                   ).
homomorph1(rule(pr(X,p(I,J),p(I2,J2)),term(_)),
                    [(rule(pr(X,p(I,J),p(I2,J2)),eps):-J2\=0),
                     (rule(pr(X,p(I,J),p(I2,J2)),term(ptrnWord(PW))):-J2=0,(I=I2,PW=I;J=0,PW=I2))]
                   ).
homomorph1(rule(X,Y,Z),[rule(X,Y,Z)]):-Y\=term(_).
homomorph1(rule(X,Y),[rule(X,Y)]):-Y\=term(_).


