/** <solveEquations> Solve equations either by using the prolog clpfd library or by translating it into yices formula. 
@author Tomas Poch
*/

%:-module(solveEquations, [generateSMT/4,solve/3]).


groundVars(_,[],[]).
groundVars(BASE,[H|T],[HV|TV]):-
    groundVar(BASE,H,HV),
    groundVars(BASE,T,TV).

groundVar(BASE,term(ptrnWord(X)),TNTV):-
    !,atom_concat(BASE,X,ATOM),ATOM=TNTV.

groundVar(BASE,TNT,TNTV):-
    term_to_atom(TNT,TNT2),atom_concat(BASE,TNT2,ATOM),ATOM=TNTV.

groundKeyValueList(_,_,[]).
groundKeyValueList(IDX,BASE,[(_-V)|T]):-
    groundVar(BASE,IDX,V),
%   groundVar(BASE,K,V),
    IDX2 is IDX+1,
    groundKeyValueList(IDX2,BASE,T).
    
groundMap(BASE,MAP):-
    assoc_to_list(MAP,MLIST),
    groundKeyValueList(0,BASE,MLIST).

groundMaps(_,_,[]).
groundMaps(BASE,IDX,[REACHTKMAP|REACHTKMAPS]):-
  IDX2 is IDX+1,
  groundMaps(BASE,IDX2,REACHTKMAPS),
  term_to_atom(IDX,AXTERM),atomconcat(BASE,AXTERM,LBASE),atomconcat(LBASE,'_',LBASE_),
  groundMap(LBASE_,REACHTKMAP).

%groundValuesByIndex(BASE,IDX,VARS)
groundByIndex(_,_,[]).
groundByIndex(BASE,IDX,[H|T]):-
  IDX2 is IDX+1,
  groundVar(BASE,IDX,H),
  groundByIndex(BASE,IDX2,T). 

 
writeEquations(TERMS,form(EQLIST,TERMVARS,RULEVARS,REACHMAPS)):-
  groundEquation(TERMS,TERMVARS,RULEVARS,REACHMAPS),
  writeList(EQLIST).

%%groundEquation(TERMS,TERMVARS,RULEVARS,REACHVARS).
%
% Ground variables to contain human/yices readable identifier. 
%
% The formula is generated using prolog variables to represent the formula variables. In order to print out/solve by yices the variables has to be ground to a string identifier.
% Even though the formula is not passed, the predicate is provided with all variables used by the formula.
%
% @param TERMS    list of terminals generated by the grammar
% @param TERMVARS list of terminal variables. Terminal variable represent number of times the terminal ocuurs in a word generated by the grammar
% @param RULEVARS list of rule variables. Rule variable represent number of times the particular rule was used when deriving a word.
% @param REACHVARS structure associating symbols with reachability variables
%
groundEquation(TERMS,TERMVARS,RULEVARS,REACHTKMAP):-
  groundVars(x_,TERMS,TERMVARS),
  (is_list(REACHTKMAP)->groundMaps(z_,0,REACHTKMAP);groundMap(z_,REACHTKMAP)),
  groundByIndex(y_,0,RULEVARS).

  
presentFormat(VARNAME,VARVALUE):-
    writef('%d = %d \t ',[VARNAME,VARVALUE]).

presentVar(BASE,TNT,TNTV):-
    atom(TNT),atom_concat(BASE,TNT,ATOM),presentFormat(ATOM,TNTV).

presentVar(BASE,TNT,TNTV):-
    integer(TNT),atom_concat(BASE,TNT,ATOM),presentFormat(ATOM,TNTV).

presentVar(BASE,term(T),TV):-
    term_to_atom(T,TT),
    atomconcat(BASE,TT,ATOM),presentFormat(ATOM,TV).

presentVars(_,[],[]).
presentVars(BASE,[H|T],[HV|TV]):-
    presentVar(BASE,H,HV),
    presentVars(BASE,T,TV).

presentKeyValueList(_,[]).
presentKeyValueList(BASE,[(K-V)|T]):-
    presentVar(BASE,K,V),
    presentKeyValueList(BASE,T).
    

presentMap(BASE,MAP):-
    assoc_to_list(MAP,MLIST),
    presentKeyValueList(BASE,MLIST).

%presentValuesByIndex(BASE,IDX,VARS)
presentByIndex(_,_,[]).
presentByIndex(BASE,IDX,[H|T]):-
  (H\==0->presentVar(BASE,IDX,H)),
  IDX2 is IDX+1,
  presentByIndex(BASE,IDX2,T). 

presentResult(TERMS,TERMVARS,RULEVARS,REACHTKMAP):-
  presentVars(x_,TERMS,TERMVARS),nl,nl,
  presentMap(z_,REACHTKMAP),nl,nl,
  presentByIndex(y_,0,RULEVARS).

printSparseArray(_,[]).
printSparseArray(IDX,[H|T]):-
  (H\==0->write(IDX->H),write(',' );true),IDX2 is IDX+1, printSparseArray(IDX2,T).


printFormulaStats(EQLIST,TERMVARS,RVARS,REACHVARMAPS):-
  length(EQLIST,NCLAUSE),
  length(TERMVARS,NTERMVARS),
  length(RVARS,NRVARS),
  maplist(assoc_to_values,REACHVARMAPS,REACHVARSLIST),
  append(REACHVARSLIST,REACHVARS),
  length(REACHVARS,NREACHVARS),
  writeln((clauses=NCLAUSE,termvars=NTERMVARS,rulevars=NRVARS,reachvars=NREACHVARS)).


clpSolveEquation(BOUND,EQ,TERMVARS,RULEVARS,REACHVARS,SOLUTIONFOUND):-
  append(RULEVARS,TERMVARS,TMP),
  append(TMP,REACHVARS,ALLVARS),
  %write((ALLVARS ins 0..BOUND,call(EQ),label(ALLVARS))),
  length(REACHVARS,REACHVARSSIZE),
  TERMVARS ins 0..BOUND,RULEVARS ins 0..BOUND,REACHVARS ins 0..REACHVARSSIZE,
  verboseWriteLn('Reification'),
  (call(EQ)->
     (verboseWriteLn('Labeling'),
      labeling([min],ALLVARS),SOLUTIONFOUND=ok;
      SOLUTIONFOUND=no-clp
     );
     SOLUTIONFOUND=no-reification
  ).

mkPrefixAtom(OP,A1,A2,R):-
    atomic_list_concat(['(',OP,' ',A1,' ',A2,')'],R).

mkPrefixAtom(OP,A1,R):-
    atomic_list_concat(['(',OP,' ',A1,')'],R).

clpfd2smtlib(T,T):-atom(T).
clpfd2smtlib(T,T):-number(T).
clpfd2smtlib(A+B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('+',A1,B1,R).
clpfd2smtlib(A-B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('-',A1,B1,R).
clpfd2smtlib(A*B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('*',A1,B1,R).
clpfd2smtlib(A#=B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('=',A1,B1,R).
clpfd2smtlib(A#\/B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('or',A1,B1,R).
clpfd2smtlib(A#/\B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('and',A1,B1,R).
clpfd2smtlib(A#>B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('>',A1,B1,R).
clpfd2smtlib(A#=<B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('=<',A1,B1,R).
clpfd2smtlib(A#==>B,R):-clpfd2smtlib(A,A1),clpfd2smtlib(B,B1),mkPrefixAtom('=>',A1,B1,R).

generateFormulas(_,_,[]).
generateFormulas(FORM,STREAM,[H|T]):-
    generateFormula(FORM,STREAM,H),
    generateFormulas(FORM,STREAM,T).


generateFormula(yices,STREAM,H):-
    clpfd2smtlib(H,F),mkPrefixAtom('assert+',F,ASSERT), write(STREAM,ASSERT),nl(STREAM).

generateFormula(smtlib2,STREAM,H):-
    clpfd2smtlib(H,F),mkPrefixAtom('assert',F,ASSERT), write(STREAM,ASSERT),nl(STREAM).



generateNatDef(yices,STREAM,H):-
    atomconcat(H,'::nat',NAMETYPE),
    mkPrefixAtom(define,NAMETYPE,DEF),
    write(STREAM,DEF),nl(STREAM).


generateNatDef(smtlib2,STREAM,H):-
    write(STREAM,'(declare-fun '),write(STREAM,H),write(STREAM,' () Int) (assert (>= '),write(STREAM,H),write(STREAM,' 0 ))'),nl(STREAM).

generateNatDefs(_,_,[]).
generateNatDefs(FORM,STREAM,[H|T]):-
    generateNatDef(FORM,STREAM,H),
    generateNatDefs(FORM,STREAM,T).

generateDefs(FORM,STREAM,TERMVARS,RULEVARS,[]):-
    generateNatDefs(FORM,STREAM,TERMVARS),
    generateNatDefs(FORM,STREAM,RULEVARS).

generateDefs(FORM,STREAM,TERMVARS,RULEVARS,[REACHMAP|T]):-
    generateDefs(FORM,STREAM,TERMVARS,RULEVARS,T),
    assoc_to_values(REACHMAP,REACHVARS),
    generateNatDefs(FORM,STREAM,REACHVARS).

printProlog(smtlib2,STREAM):-
    write(STREAM,'(set-logic QF_LIA)'),nl(STREAM),
    write(STREAM,'(set-info :smt-lib-version 2.0)'),nl(STREAM).


printProlog(yices,STREAM):-
    write(STREAM,'(set-evidence! true)'),nl(STREAM),
    write(STREAM,'(set-verbosity! 3)'),nl(STREAM).  


printEpilog(smtlib2,STREAM):-
    write(STREAM,'(check-sat)'),nl(STREAM).


printEpilog(yices,STREAM):-
    write(STREAM,'(check)'),nl(STREAM).


generateModelInspection1(yices,_,_).
generateModelInspection1(smtlib2,STREAM,H):-
  write(STREAM,'(get-value ('),write(STREAM,H),write(STREAM,') )'),nl(STREAM).
  


generateModelInspection(_,_,[]).
generateModelInspection(FORM,STREAM,[H|T]):-
  generateModelInspection1(FORM,STREAM,H),
  generateModelInspection(FORM,STREAM,T).


%% generateSMT(+FORMAT,+STREAM,+TERMS,+FORM)
%
%  Print out the given formula to the stream to be solved by an SMT solver
%
%  It involves grounding of the formula to replace prolog variables by textual variable identifiers
%  and providing proper definitions.
%
%  @param FORMAT  SMT format. Currently supports just 'yices'
%  @param STREAM  output stream to write the formula to
%  @param TERMS   list of terminals used by the grammar
%  @param FORM    formula to be printed out

generateSMT(FORM,STREAM,TERMS,form(EQLIST,TERMVARS,RULEVARS,REACHMAPS)):-
    printFormulaStats(EQLIST,TERMVARS,RULEVARS,REACHMAPS),
    printProlog(FORM,STREAM),
    groundEquation(TERMS,TERMVARS,RULEVARS,REACHMAPS),
    generateDefs(FORM,STREAM,TERMVARS,RULEVARS,REACHMAPS),
    generateFormulas(FORM,STREAM,EQLIST),
    printEpilog(FORM,STREAM),
    generateModelInspection(FORM,STREAM,TERMVARS).


%% solve(+METHOD,?FORM,-PK).
%% solve(+METHOD,?FORM,+PK).
%
% Solve the given formula by prolog.
%
% @param METHOD method used to solve the formula. Supports just 'clpfd', which requires to bound the variables. Currently, the variables are bound to 50
% @param FORM   formula to be solved
% @param PK     solution - parikh image of a word generated by the grammar.
%
solve(clpfd,form(EQLIST,TERMVARS,RULEVARS,REACHMAPS),PK):-
    conjunction(EQLIST,EQ),
    collectVarsFromMaps(REACHMAPS,REACHVARS),
    printFormulaStats(EQLIST,TERMVARS,RULEVARS,REACHMAPS),
    (nonvar(PK)->TERMVARS=PK;true),
    clpSolveEquation(50,EQ,TERMVARS,RULEVARS,REACHVARS,RESULT),
    (RESULT=ok->PK=TERMVARS;PK=RESULT).



